/* Configuration file parser
 *
 * Copyright (c) 2002, Martin Hedenfalk <mhe@home.se>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: libcfg.c,v 1.8 2002/04/17 13:44:56 mhe Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "libcfg.h"
#include <string.h>
#include <stdlib.h>

extern int yylex(cfg_t *cfg);
YYSTYPE yylval;

static cfg_errfunc_t _errfunc = 0;

cfgopt_t *cfg_getopt(cfg_t *cfg, const char *name)
{
	int i;

	if(!cfg || !cfg->opts || !name) {
		cfg_error(cfg, "cfg_getopt() called with null argument or no options");
		return 0;
	}

	for(i = 0; cfg->opts[i].name; i++) {
		if(is_set(CFGF_NOCASE, cfg->flags)) {
			if(strcasecmp(cfg->opts[i].name, name) == 0) {
				return &cfg->opts[i];
			}
		} else {
			if(strcmp(cfg->opts[i].name, name) == 0) {
				return &cfg->opts[i];
			}
		}
	}
	cfg_error(cfg, "cfg_getopt(): no such option '%s'", name);
	return 0;
}

cfgval_t *cfg_getval(cfg_t *cfg, const char *name)
{
	cfgopt_t *opt = cfg_getopt(cfg, name);
	return opt ? &opt->val : 0;
}

int cfg_getint(cfg_t *cfg, const char *name)
{
	cfgval_t *val = cfg_getval(cfg, name);
	if(val)
		return val->i;
	else
		return 0;
}

char *cfg_getstr(cfg_t *cfg, const char *name)
{
	cfgval_t *val = cfg_getval(cfg, name);
	if(val)
		return val->s;
	else
		return 0;
}

cfgval_t *cfg_sgetval(cfg_t *cfg, const char *sec, const char *name)
{
	cfgval_t *val = cfg_getval(cfg, sec);
	if(val)
		return cfg_getval(&val->sec, name);
	return 0;
}

int cfg_sgetint(cfg_t *cfg, const char *sec, const char *name)
{
	cfgval_t *val = cfg_sgetval(cfg, sec, name);
	if(val)
		return val->i;
	return 0;
}

char *cfg_sgetstr(cfg_t *cfg, const char *sec, const char *name)
{
	cfgval_t *val = cfg_sgetval(cfg, sec, name);
	if(val)
		return val->s;
	return 0;
}

cfg_t *cfg_tgetsec(cfg_t *cfg, const char *secname, const char *title)
{
	int i;
	cfgopt_t *opt;

	opt = cfg_getopt(cfg, secname);
	if(opt) {
		if(opt->type != CFGT_SEC || !is_set(CFGF_TITLE, opt->flags))
			return 0;

		if(is_set(CFGF_MULTI, opt->flags)) {
			for(i = 0; i < opt->val.a.size; i++) {
				if(strcmp(title, opt->val.a.val[i].sec.title) == 0)
					return &opt->val.a.val[i].sec;
			}
		} else {
			if(strcmp(title, opt->val.sec.title) == 0)
				return &opt->val.sec;
		}
	}
	cfg_error(cfg, "cfg_tgetsec(): no such section '%s'", title);
	return 0;
}

cfgval_t *cfg_stgetval(cfg_t *cfg, const char *sec,
					   const char *title, const char *name)
{
	cfg_t *_sec;

	_sec = cfg_tgetsec(cfg, sec, title);
	if(_sec)
		return cfg_getval(_sec, name);
	return 0;
}

int cfg_stgetint(cfg_t *cfg, const char *sec, const char *title,
				 const char *name)
{
	cfgval_t *val = cfg_stgetval(cfg, sec, title, name);
	if(val)
		return val->i;
	return 0;
}

char *cfg_stgetstr(cfg_t *cfg, const char *sec, const char *title,
				   const char *name)
{
	cfgval_t *val = cfg_stgetval(cfg, sec, title, name);
	if(val)
		return val->s;
	return 0;
}

cfgval_t *cfg_addval(cfgval_t *v, cfgopt_t *o)
{
	cfgarray_t *a = &v->a;

	if(a->size)
		a->val = (cfgval_t *)realloc(a->val,
									 (a->size + 1) * sizeof(cfgval_t));
	else
		a->val = (cfgval_t *)malloc(sizeof(cfgval_t));
	return &a->val[a->size++];
}

static cfgopt_t *cfg_dupopts(cfgopt_t *opts)
{
	int n;
	cfgopt_t *dupopts;

	for(n = 0; opts[n].name; n++)
		/* do nothing */ ;
	dupopts = (cfgopt_t *)malloc(++n * sizeof(cfgopt_t));
	memcpy(dupopts, opts, n * sizeof(cfgopt_t));
	return dupopts;
}

static void cfg_init_defaults(cfgopt_t *opts)
{
	int i;

	for(i = 0; opts[i].name; i++) {
		if(is_set(CFGF_LIST, opts[i].flags)) {
			/* FIXME: setting defaults for lists are not yet supported
			 */
			continue;
		}
		switch(opts[i].type) {
			case CFGT_INT:
				opts[i].val.i = (int)opts[i].def;
				break;
			case CFGT_STR:
				if(opts[i].def)
					opts[i].val.s = strdup((char *)opts[i].def);
				break;
			case CFGT_SEC:
				break;
			case CFGT_NONE:
			case CFGT_BOOL:
			default:
				break;
		}
	}
}

cfgval_t *cfg_setopt(cfg_t *cfg, const char *name, void *value)
{
	cfgval_t *val;
	cfgopt_t *opt = cfg_getopt(cfg, name);

	if(opt == 0) {
		cfg_error(cfg, "unrecognized option '%s'", name);
		return 0;
	}

	if(opt->callback)
		(*opt->callback)(cfg, name, value);
	/* FIXME: terminate function depending on return value from callback */

	if(is_set(CFGF_MULTI, opt->flags) || is_set(CFGF_LIST, opt->flags))
		val = cfg_addval(&opt->val, opt);
	else
		val = &opt->val;

	switch(opt->type) {
		case CFGT_INT:
			val->i = (int)value;
			break;
		case CFGT_STR:
			if(val->s)
				free(val->s);
			val->s = strdup((char *)value);
			break;
		case CFGT_SEC:
			val->sec.name = strdup(opt->name);
			val->sec.title = 0;
			val->sec.opts = cfg_dupopts(opt->subopts);
			cfg_init_defaults(val->sec.opts);
			val->sec.flags = 0;
			val->sec.filename = cfg->filename;
			val->sec.line = cfg->line;
			break;
		default:
			cfg_error(cfg, "internal error in cfg_setopt()");
			break;
	}
	return val;
}

cfg_errfunc_t cfg_set_error_function(cfg_errfunc_t errfunc)
{
	cfg_errfunc_t old = _errfunc;
	_errfunc = errfunc;
	return old;
}

void cfg_error(cfg_t *cfg, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);

	if(_errfunc) {
		(*_errfunc)(cfg, fmt, ap);
	} else {
		if(cfg && cfg->filename && cfg->line)
			fprintf(stderr, "%s:%d: ", cfg->filename, cfg->line);
		else if(cfg && cfg->filename)
			fprintf(stderr, "%s: ", cfg->filename);
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	}

	va_end(ap);

	exit(4);
}

cfg_t *cfg_init(cfgopt_t *opts, int flags)
{
	cfg_t *cfg;

	cfg = (cfg_t *)malloc(sizeof(cfg_t));
	cfg->name = "global";
	cfg->opts = opts;
	cfg->flags = flags;
	cfg->filename = 0;
	cfg->line = 0;

	cfg_init_defaults(cfg->opts);

	return cfg;
}

static int cfg_parse_internal(cfg_t *cfg, int level)
{
	int state = 0;
	char *optname = 0;
	char *opttitle = 0;
	cfgopt_t *opt = 0;
	cfgval_t *val;

	while(1) {
		int tok = yylex(cfg);
		int r;

		if(tok == EOF) {
			if(state != 0)
				cfg_error(cfg, "premature end of file");
			return 1;
		}

		switch(state) {
			case 0: /* expecting a option name */
				if(tok == '}') {
					if(level == 0)
						cfg_error(cfg, "unexpected closing brace");
					return 0;
				}
				if(tok != CFGT_STR)
					cfg_error(cfg, "expected option name");
				optname = strdup(yylval.s);
				opt = cfg_getopt(cfg, optname);
				if(opt == 0)
					cfg_error(cfg, "unrecognized option '%s'", optname);
				if(opt->type == CFGT_SEC) {
					if(is_set(CFGF_TITLE, opt->flags))
						state = 7;
					else
						state = 6;
				} else
					state = 1;
				break;
			case 1: /* expecting an equal sign */
				if(tok != '=')
					cfg_error(cfg, "missing equal sign after option");
				if(is_set(CFGF_LIST, opt->flags))
					state = 3;
				else
					state = 2;
				break;
			case 2: /* expecting an option value */
				/* FIXME: what about BOOLs? (1/0/true/false/on/off/yes/no */
				if(opt->type != tok)
					cfg_error(cfg, "mismatched value type");
				cfg_setopt(cfg, optname, (void *)yylval.s);
				state = 0;
				break;
			case 3: /* expecting an opening brace for a list option */
				if(tok != '{')
					cfg_error(cfg, "missing opening brace for list %s",
							  optname);
				state = 4;
				break;
			case 4: /* expecting an option value for a list option */
				/* FIXME: what about BOOLs? (1/0/true/false/on/off/yes/no */
				if(opt->type != tok)
					cfg_error(cfg, "mismatched value type");
				cfg_setopt(cfg, optname, (void *)yylval.s);
				state = 5;
				break;
			case 5: /* expecting a separator for a list option, or
					 * closing (list) brace */
				if(tok == ',')
					state = 4;
				else if(tok == '}')
					state = 0;
				else
					cfg_error(cfg, "unexpected token");
				break;
			case 6: /* expecting an opening brace for a section */
				if(tok != '{') {
					cfg_error(cfg, "missing opening brace for section %s",
							  optname);
				}

				val = cfg_setopt(cfg, optname, 0);
				if(opttitle) {
					val->sec.title = opttitle;
					opttitle = 0;
				}

				if(is_set(CFGF_MULTI, opt->flags))
					r = cfg_parse_internal(
						&opt->val.a.val[opt->val.a.size-1].sec, level+1);
				else
					r = cfg_parse_internal(&opt->val.sec, level+1);
				if(r == 1) {
					return 1;
				}
				state = 0;
				break;
			case 7: /* expecting a title for a section */
				if(tok != CFGT_STR) {
					cfg_error(cfg, "missing title for section %s", optname);
				} else {
					opttitle = strdup(yylval.s);
				}
				state = 6;
				break;
		}
	}
}

int cfg_parse(cfg_t *cfg, const char *filename)
{
	cfg->filename = filename;
	cfg->line = 1;

	yyin = fopen(filename, "r");
	if(yyin == 0) {
		cfg_error(cfg, "unable to open configuration file");
		exit(6);
	}

	cfg_parse_internal(cfg, 0);
	return 0;
}

void cfg_free(cfg_t *cfg)
{
}
