/* Open.Net Authentication Daemon
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: common.c,v 1.18 2002/04/25 12:03:32 mhe Exp $
 */

#include "common.h"

int use_syslog = 0;
int use_debug = 0;
char *program_name = PACKAGE;

/* Return the integer code for a logging facility. The names are
 * specified in <syslog.h>.
 */
#define DEFLOG(fac) {fac, #fac}
int log_facility(char *str)
{
	struct {
		int facility;
		char *name;
	} faclist[] = {
		DEFLOG(LOG_AUTHPRIV), DEFLOG(LOG_DAEMON), DEFLOG(LOG_LOCAL0),
		DEFLOG(LOG_LOCAL1), DEFLOG(LOG_LOCAL2), DEFLOG(LOG_LOCAL3),
		DEFLOG(LOG_LOCAL4), DEFLOG(LOG_LOCAL5), DEFLOG(LOG_LOCAL6),
		DEFLOG(LOG_LOCAL7), DEFLOG(LOG_USER), {0, 0}
	};
	int i;

	if(str == 0)
		return LOG_USER;

	for(i=0; faclist[i].name; i++) {
		if(strcmp(str, faclist[i].name) == 0)
			return faclist[i].facility;
	}
	return LOG_USER;
}

/* Logs a message. If the global variable use_syslog is set, the
 * message is sent to the system log (facility configured with
 * log-facility), otherwise print directly to the terminal (on
 * stderr). The global variable debug specifies if debug messages
 * should be ignored or not.
 */
void vlogmsg(int priority, const char *format, va_list ap)
{
	static int fac = -1;

	if(fac == -1)
		fac = log_facility(cfg_getstr(cfg, "log-facility"));

	if(use_syslog) {
#ifdef HAVE_VSYSLOG
		vsyslog(priority | fac, format, ap);
#else
		char *tmp;
		vasprintf(&tmp, format, ap);
		syslog(priority | fac, tmp);
		free(tmp);
#endif
	} else {
		char tbuf[128];
		time_t now;
		time(&now);
		strftime(tbuf, 128, "%H:%M:%S", localtime(&now));
		fprintf(stderr, "%s %s[%ld/%d]: ", tbuf, program_name, pthread_self(),
				getpid());
#ifdef LOGMSG_COLOR
		if(priority == LOG_ERR)
			fprintf(stderr, "\e[31m");
		else if(priority == LOG_INFO)
			fprintf(stderr, "\e[32m");
		else if(priority == LOG_WARNING)
			fprintf(stderr, "\e[36m");
#endif
		vfprintf(stderr, format, ap);
#ifdef LOGMSG_COLOR
		fprintf(stderr, "\e[0m\n");
#endif
		fflush(stderr);
	}
}

void logmsg(int priority, const char *format, ...)
{
	va_list ap;

	if(priority == LOG_DEBUG && !use_debug)
		return;

	va_start(ap, format);
	vlogmsg(priority, format, ap);
	va_end(ap);
}

void logmsg_cfg(cfg_t *cfg, const char *fmt, va_list ap)
{
	vlogmsg(LOG_ERR, fmt, ap);
}

/* mapping of status return codes to strings
 *
 * don't change this unless you also change req_status_t in common.h
 */
static const char * const sys_statuslist[] = {
	"Authentication successful",
	"Authentication failed",
	"Invalid request",
	"Bad request",
	"Domain does not exist",
	"Failed to open rule in firewall",
	"Authentication backend failed",
	"Domain has reached maximum number of allowed users",
	"User already exists",
	"User does not exist",
	"Connection to firewall control daemon failed",
	"The firewall control daemon failed for some reason"
};

/* Returns a string describing the status code. Returns a statically
 * allocated buffer, overwritten on each call.
 */
const char *strstatus(req_status_t status)
{
	if(status >= 0 && status < sizeof(sys_statuslist)/sizeof(char *))
		return sys_statuslist[status];
	return "Unknown status code";
}

/* Returns a string describing the MAC address. Returns a statically
 * allocated buffer, overwritten on each call.
 */
char *strmac(unsigned char mac[])
{
	static char macstr[18];

	sprintf(macstr, "%02x:%02x:%02x:%02x:%02x:%02x",
			mac[0], mac[1], mac[2],
			mac[3], mac[4], mac[5]);
	return macstr;
}

/* trims leading and trailing whitespace from STR
 */
char *trim(char *str)
{
	char *e;

	while(isspace(*str))
		++str;
	e = str + strlen(str);
	while(--e != str && isspace(*e))
		/* do nothing */ ;
	*(e+1) = 0;

	return str;
}

char *strmstatus(mstatus_t ms)
{
	switch(ms) {
		case MS_OK:
			return "OK";
		case MS_PROBING:
			return "PROBING";
		case MS_UPDATING:
			return "UPDATING";
		case MS_LOGOUT:
			return "LOGOUT";
		case MS_DETACHED:
			return "DETACHED";
		default:
			return "GAZONK!?";
	}
}

void block_term_sigs(void)
{
	sigset_t sigmask;

	/* Block termination signals, they are handled by the main thread
	 */
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGINT);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGHUP);
	sigaddset(&sigmask, SIGUSR1);
	pthread_sigmask(SIG_SETMASK, &sigmask, 0);
}
