/* Open.Net Authentication Daemon
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: common.h,v 1.27 2002/04/19 11:51:15 mhe Exp $
 */

#ifndef _debug_h_
#define _debug_h_

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include <libcfg.h>

#include "list.h"

#include <openssl/ssl.h>

#ifdef HAVE_LIBDMALLOC
# include <dmalloc.h>
#endif

#include <sys/types.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#include <pthread.h>

#if defined(HAVE_DECL_VASPRINTF) && !HAVE_DECL_VASPRINTF
int vasprintf(char **strp, const char *format, va_list ap);
#endif

#if defined(HAVE_DECL_ASPRINTF) && !HAVE_DECL_ASPRINTF
int asprintf(char **strp, const char *format, ...);
#endif

#define MAX_MAC_LEN 6

enum req_cmd_t {REQ_CMD_LOGIN, REQ_CMD_LOGOUT, REQ_CMD_STATS, REQ_CMD_LIST};
typedef enum req_cmd_t req_cmd_t;

/* status return codes
 *
 * Don't change this unless you also change sys_statuslist in
 * common.c
 */
enum req_ret_t {REQ_RET_OK, REQ_RET_FAIL};
enum req_status_t {
	REQ_STATUS_OK,        /* ok, user is authenticated */
	REQ_STATUS_FAIL,      /* user failed authentication */
	REQ_STATUS_NOCMD,     /* invalid request command */
	REQ_STATUS_BADREQ,    /* bad request (malformed parameters etc...) */
	REQ_STATUS_NODOMAIN,  /* domain does not exist */
	REQ_STATUS_FWFAILED,  /* failed to open rule in firewall */
	REQ_STATUS_BACKFAIL,  /* the authentication backend failed */
	REQ_STATUS_REJFULL,   /* user was rejected because domain is full */
	REQ_STATUS_USEREXIST, /* user is already logged in */
	REQ_STATUS_NOUSER,    /* no such user */
	REQ_STATUS_FWCONNFAIL,/* connection to fwcd failed */
	REQ_STATUS_FWCDFAIL   /* fwcd failed for some reason */
};
typedef enum req_ret_t req_ret_t;
typedef enum req_status_t req_status_t;

/* monitoring status
 */
enum mstatus_t {
	MS_OK,          /* normal state, ok to probe */
	MS_PROBING,     /* user is currently being probed */
	MS_UPDATING,    /* updating missed probe counter */
	MS_LOGOUT,      /* user is being logged out */
	MS_DETACHED     /* user has been logged out and memory should be freed */
};
typedef enum mstatus_t mstatus_t;

enum fwtype_t { FW_LOCAL, FW_REMOTE };
typedef enum fwtype_t fwtype_t;

typedef struct userdata_t userdata_t;

struct domain_t {
	char *name; /* name of this domain */
	fwtype_t fwtype; /* local or remote control of firewall? */
	unsigned int nusers; /* number of users currently active on this domain */
	/* handle to probe library if fwtype == FW_LOCAL, or 0 if no
	 * probing should be performed */
	void *probelib_handle;
	int (*probe_user_func)(userdata_t *, int);
	void (*probe_cleanup_func)(userdata_t *);

	SSL *sslout; /* connection to firewall, if fw_type == FW_REMOTE */
	SSL *sslin;  /* connection from firewall, if fw_type == FW_REMOTE */
};
typedef struct domain_t domain_t;

/* data stored for each logged in user in the monitoring list
 */
struct userdata_t {
	unsigned char mac[MAX_MAC_LEN];   /* 00:02:44:16:E9:71 */
	struct in_addr ip;                /* 32 bits unsigned integer */
	char *username;                   /* username, doh! */
	domain_t *domain;                 /* pointer to domain struct */
	int missed_probes;                /* number of missed probes so far */
	int failed_logouts;               /* number of failed logout attempts */
	mstatus_t mstatus;                /* monitoring status */
	pthread_t probe_id;               /* id of thread probing this user */
	struct timeval tv;                /* probe timeout */
	void *probe_data;                 /* data specific for the probe thread */
	pthread_mutex_t mutex;
	struct timeval logintime;         /* time user logged in */
};

/* main.c
 */
extern cfg_t *cfg;
extern int num_domain;

/* common.c
 */
extern int use_debug;
extern int use_syslog;
extern char *program_name;
int spawn(char * const argv[]);
int log_facility(char *str);
const char *strstatus(req_status_t status);
void vlogmsg(int priority, const char *format, va_list ap);
void logmsg(int priority, const char *format, ...);
void logmsg_cfg(cfg_t *cfg, const char *fmt, va_list ap);
char *strmac(unsigned char mac[]);
char *trim(char *str);
char *strmstatus(mstatus_t ms);
void block_term_sigs(void);

/* request.c
 */
void request_loop(void);
extern unsigned long int failed_requests;
extern unsigned long int failed_auths;
extern unsigned long int total_logins;
extern time_t oasis_start;

/* auth.c
 */
req_status_t login(userdata_t *ud, char *password);
req_status_t logout(userdata_t *ud);
void reset_firewalls(void);

/* monitor.c
 */
extern domain_t *domains;
extern list_t *users; /* list of userdata_t */
extern list_t *logout_users; /* list of userdata_t */
extern pthread_mutex_t modify_list_mutex;
extern pthread_mutex_t logout_users_mutex;

void userdata_destroy(userdata_t *ud);
userdata_t *userdata_create(char *username, char *domain,
							char *password, struct in_addr *ip,
							unsigned char mac[]);
req_status_t monitor_check_user(userdata_t *ud);
req_status_t monitor_insert(userdata_t *req);
req_status_t monitor_remove(userdata_t *req);
void monitor_shutdown(void);
void monitor_init(void);
userdata_t *search_user(const char *username, const char *domain);
userdata_t *search_ipmac(struct in_addr *ip, unsigned char mac[6]);
mstatus_t get_mstatus(userdata_t *ud);
mstatus_t set_mstatus(userdata_t *ud, mstatus_t mstatus);
int get_domain_size(domain_t *domain, req_status_t *status);

/* stats.c
 */
req_status_t stats_send(FILE *fp);
req_status_t stats_send_domain(FILE *fp, char *domain);
req_status_t stats_send_user(FILE *fp, char *username, char *domain);

/* fw.c
 */
req_status_t fw_open(userdata_t *ud);
req_status_t fw_close(userdata_t *ud);
void fw_reset(domain_t *domain);

/* The authentication function is responsible of the actual
 * authentication of the user. This function can use whatever backend it
 * wants. Currently it uses PAM.
 */
req_status_t authenticate(char *username, char *domain, char *password);


extern int experimental;

#endif
