/* Firewall Control Daemon
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: fwcd.c,v 1.21 2002/05/02 09:13:53 mhe Exp $
 */

/* TODO:
 * - Will this scale to 3000+ users?
 * - Consider getting packet counts from iptables instead
 * - Write a cleaner hash table
 * - Clean up this mess :-) ...some comments might be good
 */

#include "common.h"
#include "xmalloc.h"

#include <libnet.h>
#include <pcap.h>
#include <sys/time.h>
#include <pthread.h>
#include <signal.h>
#include <dlfcn.h>

#include "fwssl.h"
#include "llhash.h"

#include <libcfg.h>

cfg_t *cfg = 0;
void *probelib_handle = 0;
int (*probe_user_func)(userdata_t *, int) = 0;
void (*probe_cleanup_func)(userdata_t *) = 0;

#define _GNU_SOURCE
#include <getopt.h>

LLHTABLE hashtable;
SSL *ssl_as = 0; /* SSL connection to the access server */
pthread_t probe_thread;
pthread_t packet_check_thread;

pthread_mutex_t hashmutex = PTHREAD_MUTEX_INITIALIZER;

req_status_t fwcd_close(struct in_addr *ip, unsigned char mac[6],
						int force_removal);

#define COPYLINE "fwcd version " VERSION "\n" \
"Copyright (c) 2002, Royal Institute of Technology. All rights reserved.\n\n"

#define SYNTAX \
"Usage: fwcd [options]\n" \
"Options:\n" \
"  -f, --fg            run in foreground (don't fork, send messages to terminal)\n" \
"  -d, --debug         include debug messages in syslog\n" \
"  -r, --rcfile=FILE   read configuration from FILE\n" \
"  -h, --help          show this help and exit successfully\n" \
"  -V, --version       show version information and exit successfully\n" \
"\n" \
"Report bugs to <open-devel@stockholmopen.net>\n" \
"\n"

int makekey(unsigned char mac[6])
{
	int key = 0;
	int i;

	for(i=0; i<6; i++)
		key = (key ^ mac[i]) + ((key<<26)+(key>>6));
/*		key += mac[i];*/

	return key & 0x7FFFFF;
}

int hashfunc(int key)
{
	return key;
}

void *packet_check_thread_main(void *arg)
{
	int i;

	block_term_sigs();

	while(1) {
		sleep(cfg_getint(cfg, "probe-interval"));

		pthread_mutex_lock(&hashmutex);
/*		logmsg(LOG_DEBUG, "checking packet counts for %d users...",
		llhsize(hashtable));*/
		for(i = 0; i < LLHSLOTS; ++i) {
			struct item **item = &hashtable[i];
			struct item *kill;

			while(*item) {
				logmsg(LOG_DEBUG, "checking %s (%d/%d)",
					   strmac((*item)->d->mac), (*item)->d->packets,
					   (*item)->d->misses);
				if((*item)->d->packets == 0) {
					if(++(*item)->d->misses > cfg_getint(cfg,
														 "max-missed-probes"))
					{
						kill = *item;
						if(fwcd_close(&kill->d->ip, kill->d->mac,
									  kill->d->failed_logouts > 3 ? 1 : 0)
						   == REQ_STATUS_OK)
						{
							*item = (*item)->next;
							free(kill->d);
							free(kill);
						} else
							(*item)->d->failed_logouts++;
					} else if(probe_user_func) {
						int ret;

						/* Probe the user by calling the dynamically
						 * loaded probe function, no need to wait for
						 * a response, it will be received by the
						 * packet checking thread
						 */
						userdata_t *ud;

						ud = xmalloc(sizeof(userdata_t));
						memcpy(ud->mac, (*item)->d->mac, MAX_MAC_LEN);
						ud->ip = (*item)->d->ip;
						ud->username = 0;
						ud->domain = 0;
						ud->mstatus = MS_OK;
						ud->missed_probes = (*item)->d->misses;
						ud->failed_logouts = 0;
						ud->probe_id = 0;
						timerclear(&ud->tv);
						ud->probe_data = (void *)0xDEADBEEF;

						logmsg(LOG_DEBUG, "probing  %s (%s)",
							   strmac(ud->mac), inet_ntoa(ud->ip));
						ret = (*probe_user_func)(ud, 0);
						if(ret != 0)
							logmsg(LOG_DEBUG,
								   "probe plugin returned with exit"
								   " code %d (ignored)", ret);
						xfree(ud);
					}
				} else {
					(*item)->d->misses = 0;
					(*item)->d->packets = 0;
				}
				if(*item)
					item = &((*item)->next);
			}
		}
		pthread_mutex_unlock(&hashmutex);
	}
}

static void packet_callback(void *userdata,
							struct pcap_pkthdr *h,
							u_char *packet)
{
	struct libnet_ethernet_hdr *eth = (struct libnet_ethernet_hdr *)packet;
	u_char *payload = packet + sizeof(struct libnet_ethernet_hdr);
	struct in_addr ipaddr, destaddr;
	LLHITEM *item;
	char tmp[128];

	if(htons(eth->ether_type) == ETHERTYPE_ARP) {
		ipaddr.s_addr = *((long *)((struct libnet_arp_hdr *)payload)->ar_spa);
	destaddr.s_addr = *((long *)((struct libnet_arp_hdr *)payload)->ar_tpa);
	} else if(htons(eth->ether_type) == ETHERTYPE_IP) {
		ipaddr.s_addr = ((struct libnet_ip_hdr *)payload)->ip_src.s_addr;
		destaddr.s_addr = ((struct libnet_ip_hdr *)payload)->ip_dst.s_addr;
	} else
		logmsg(LOG_ERR, "strange packet, ether_type == %d", eth->ether_type);

	strcpy(tmp, inet_ntoa(destaddr));
#if 0
	logmsg(LOG_DEBUG, "got %s packet from %s (%s) to %s",
		   htons(eth->ether_type) == ETHERTYPE_ARP ? "ARP" : "IP",
		   strmac(eth->ether_shost), inet_ntoa(ipaddr), tmp);
#endif

	pthread_mutex_lock(&hashmutex);
	item = llhlookup(hashtable, hashfunc, makekey(eth->ether_shost));
	if(item && (*item)->ip.s_addr == ipaddr.s_addr)
		(*item)->packets++;
	pthread_mutex_unlock(&hashmutex);
}

void *probe_thread_main(pcap_t *pcap)
{
	block_term_sigs();

	while(1) {
		if(pcap_loop(pcap, 1, (pcap_handler)packet_callback, 0) == -1)
			printf("pcap_loop(): error\n");
	}
}

int probe_thread_init(void)
{
	pcap_t *pcap;
	char errbuf[LIBNET_ERRBUF_SIZE];
	char *ifname = cfg_getstr(cfg, "probe-interface");
	struct bpf_program bp;
	char *filter = 0;

	logmsg(LOG_INFO, "probe thread starting on interface %s...", ifname);

	/*
	 * pcap init
	 */
	if(!(pcap = pcap_open_live(ifname, sizeof(struct libnet_ethernet_hdr) +
							   sizeof(struct libnet_ip_hdr),
							   cfg_getint(cfg, "promiscuous"), 10, errbuf)))
	{
		printf("pcap_open_live(): %s\n", errbuf);
		return 1;
	}

	asprintf(&filter, "( ip or ip6 or arp ) and src net %s",
			 cfg_getstr(cfg, "network"));
	logmsg(LOG_DEBUG, "filter string: %s", filter);
	if(pcap_compile(pcap, &bp, filter, 0, -1) == -1) {
		printf("pcap_compile(): %s\n", pcap_geterr(pcap));
		return 1;
	}

	if(pcap_setfilter(pcap, &bp) == -1) {
		printf("pcap_setfilter(): error\n");
		return 1;
	}

	llhinit(hashtable);

	pthread_create(&packet_check_thread, 0,
				   (void *(*)(void *))packet_check_thread_main, 0);

	pthread_create(&probe_thread, 0,
				   (void *(*)(void *))probe_thread_main, pcap);

	return 0;
}

/* open a rule for a specified user
 */
req_status_t fwcd_open(fwreq_t *req)
{
	struct traffic *i;
	req_status_t status;
	int j;

	status = fw_open_local(cfg_getstr(cfg, "firewall-program"),
						   &req->ip, req->mac);

	if(status != REQ_STATUS_OK)
		return status;
	
	i = malloc(sizeof(struct traffic));
	memcpy(&i->ip, &req->ip, sizeof(struct in_addr));
	memcpy(i->mac, req->mac, 6);
	i->packets = 0;
	i->misses = 0;
	i->key = makekey(i->mac);
	i->failed_logouts = 0;

	logmsg(LOG_INFO, "open %s %s", inet_ntoa(req->ip), strmac(req->mac));
	for(j = 0; j < 100; j++) {
		int r;

		pthread_mutex_lock(&hashmutex);
		r = llhinsert(hashtable, hashfunc, i);
		pthread_mutex_unlock(&hashmutex);
		if(r == -1) {
			status = REQ_STATUS_FWCDFAIL;
			i->key++;
		} else if(r == 1) {
			status = REQ_STATUS_FWCDFAIL;
			break;
		} else {
			status = REQ_STATUS_OK;
			break;
		}
	}

	if(status == REQ_STATUS_FWCDFAIL)
		logmsg(LOG_ERR, "FAILED TO INSERT %s %s, key == %u",
			   inet_ntoa(req->ip), strmac(req->mac), i->key);

	return status;
}

static void check_oasis_connection(void)
{
	if(ssl_as == 0) {
		SSL_CTX *ctx_as = init_ctx(cfg_getstr(cfg, "certificate-file"),
								   cfg_getstr(cfg, "key-file"),
								   cfg_getstr(cfg, "oasis-certificate-file"),
								   1);
		if(ctx_as) {
			ssl_as = init_ssl_client(ctx_as,
									 cfg_getstr(cfg, "oasis-host"),
									 cfg_getint(cfg, "oasis-port"));
		}
	}
}

/* the mutex hashmutex is assumed to be already locked
 */
static req_status_t fwcd_logout(struct in_addr *ip, unsigned char mac[6])
{
	req_status_t status;

	logmsg(LOG_INFO, "close %s %s", inet_ntoa(*ip), strmac(mac));
	status = fw_close_local(cfg_getstr(cfg, "firewall-program"), ip, mac);
	logmsg(LOG_DEBUG, "fw_close_local(%s, %s) returned status '%s'",
		   inet_ntoa(*ip), strmac(mac), strstatus(status));

#if 0
	if(status == REQ_STATUS_OK)
		llhremove(hashtable, hashfunc, makekey(mac));
#endif

	return status;
}

/* close the firewall rule for a specified user, the mutex hashmutex
 * is assumed to be already locked
 */
req_status_t fwcd_close(struct in_addr *ip, unsigned char mac[6],
						int force_removal)
{
	fwreq_t req;
	fwreply_t reply;
	int ret;
	req_status_t status;

	if(!force_removal) {
		status = fwcd_logout(ip, mac);
		if(status != REQ_STATUS_OK)
			return status;
	}

	req.cmd = FWCMD_CLOSE;
	memcpy(&req.ip, ip, sizeof(struct in_addr));
	memcpy(req.mac, mac, 6);

	check_oasis_connection();

	ret = SSL_write(ssl_as, &req, sizeof(fwreq_t));
	if(check_ssl_io(ret, ssl_as, sizeof(fwreq_t)) != 0)
		return REQ_STATUS_FAIL;

	ret = SSL_read(ssl_as, &reply, sizeof(fwreply_t));
	if(check_ssl_io(ret, ssl_as, sizeof(fwreply_t)) != 0)
		return REQ_STATUS_FAIL;

	return REQ_STATUS_OK;
}

/* resets all configured firewalls
 */
req_status_t fwcd_reset(void)
{
	logmsg(LOG_INFO, "reset");

	fw_reset_local(cfg_getstr(cfg, "firewall-program"));
	/* ignore return value
	 */

	pthread_mutex_lock(&hashmutex);
	llhclear(hashtable);
	pthread_mutex_unlock(&hashmutex);
	return REQ_STATUS_OK;
}

int fwcd_process_request(SSL *ssl)
{
	char buf[1024];
	int ret;
	fwreq_t *req;
	fwreply_t reply;

	ret = SSL_read(ssl, buf, 1024);

	if(ret == 0) {
		logmsg(LOG_WARNING, "connection closed by peer");
		return -2;
	}

	if(ret < 0) {
		logmsg_ssl(LOG_ERR);
		return -1;
	}

	if(ret != sizeof(fwreq_t)) {
		logmsg(LOG_ERR, "received request too small");
		return -1;
	}

	req = (fwreq_t *)buf;

	switch(req->cmd) {
		case FWCMD_OPEN:
			reply.status = fwcd_open(req);
			break;
		case FWCMD_CLOSE:
			pthread_mutex_lock(&hashmutex);
			reply.status = fwcd_logout(&req->ip, req->mac);
			if(reply.status == REQ_STATUS_OK)
				llhremove(hashtable, hashfunc, makekey(req->mac));
			pthread_mutex_unlock(&hashmutex);
			break;
		case FWCMD_RESET:
			reply.status = fwcd_reset();
			break;
		case FWCMD_STATS:
		case FWCMD_SYNC_USER:
		default:
			logmsg(LOG_WARNING, "unknown request code: %d", req->cmd);
			reply.status = REQ_STATUS_NOCMD;
			break;
	}

	ret = SSL_write(ssl, &reply, sizeof(fwreply_t));
	if(ret <= 0) {
		logmsg_ssl(LOG_ERR);
		return -1;
	}

	return 0;
}

void cleanup(void)
{
	fwcd_reset();
#if 0
	logmsg(LOG_INFO, "shutting down packet checking threads...");
	if(pthread_cancel(probe_thread) != 0)
		logmsg(LOG_ERR, "failed to cancel probe thread");
	/* wait for the thread to exit */
	pthread_join(probe_thread, 0);

	if(pthread_cancel(packet_check_thread) != 0)
		logmsg(LOG_ERR, "failed to cancel packet check thread");
	pthread_join(packet_check_thread, 0);
#endif
}

void sig_term_handler(int signo)
{
	logmsg(LOG_INFO, "Caught signal %d, shutting down...", signo);
	cleanup();

	/* Now reraise the signal.  We reactivate the signal's default
	 * handling, which is to terminate the process. We could just
	 * call `exit' or `abort', but reraising the signal sets the
	 * return status from the process correctly.
	 */
	signal(signo, SIG_DFL);
	raise(signo);
}


void sigusr1_handler(int signo)
{
	use_debug = !use_debug;
	logmsg(LOG_INFO, "Debugging is now turned %s", use_debug ? "on" : "off");
	signal(SIGUSR1, sigusr1_handler);
}

/* Setup signal handlers
 */
void install_signal_handlers(void)
{
	struct sigaction action;
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, SIGINT);
	sigaddset(&mask, SIGQUIT);
	action.sa_handler = sig_term_handler;
	action.sa_mask = mask;
	action.sa_flags = 0;
	sigaction(SIGTERM, &action, NULL);

	sigemptyset(&mask);
	sigaddset(&mask, SIGTERM);
	sigaddset(&mask, SIGQUIT);
	action.sa_handler = sig_term_handler;
	action.sa_mask = mask;
	action.sa_flags = 0;
	sigaction(SIGINT, &action, NULL);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGUSR1, sigusr1_handler);
}

int you_can_dance_now(void)
{
	char *probe_lib;
	SSL_CTX *ctx;

	probe_lib = cfg_getstr(cfg, "probe-library");
	if(probe_lib && probe_lib[0]) {
		probelib_handle = dlopen(probe_lib, RTLD_NOW);
		if(probelib_handle == 0) {
			fprintf(stderr, "%s\n", dlerror());
			return 3;
		} else {
			logmsg(LOG_INFO, "registered probe library %s", probe_lib);
		}
		probe_user_func = dlsym(probelib_handle, "probe_user");
		if(probe_user_func == 0) {
			fprintf(stderr, "probe_user(): %s\n", dlerror());
			return 3;
		}
		probe_cleanup_func = dlsym(probelib_handle, "probe_cleanup");
	} else
		logmsg(LOG_INFO, "no probe-library defined, probing disabled");

	if(probe_thread_init() != 0)
		return 1;

	ctx = init_ctx(cfg_getstr(cfg, "certificate-file"),
				   cfg_getstr(cfg, "key-file"),
				   cfg_getstr(cfg, "oasis-certificate-file"),
				   0);
	if(!ctx)
		return 1;

	while(1) {
		int ret = init_ssl_daemon(ctx, cfg_getint(cfg, "port"),
								  fwcd_process_request);
		if(ssl_as) {
			logmsg(LOG_WARNING, "Shutting down oasis connection");
			SSL_shutdown(ssl_as);
			ssl_as = 0;
		}
		if(ret == 0)
			fwcd_reset();
	}

	return 0;
}

int main(int argc, char **argv)
{
	pid_t pid;
	struct option longopts[] = {
		{"fg", no_argument, 0, 'f'},
		{"debug", no_argument, 0, 'd'},
		{"rcfile", required_argument, 0, 'r'},
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'V'},
		{0, 0, 0, 0}
	};
	int c;
	int foreground = 0; /* run in foreground (don't fork) */
	char *rcfile = SYSCONFDIR "/fwcd.conf";
	cfgopt_t opts[] = {
		CFG_STR("certificate-file", SYSCONFDIR "/fwcd.cert", 0),
		CFG_STR("key-file", SYSCONFDIR "/fwcd.key", 0),
		CFG_STR("oasis-certificate-file", SYSCONFDIR "/oasis.cert", 0),
		CFG_INT("probe-interval", 5, 0),
		CFG_INT("max-missed-probes", 3, 0),
		CFG_STR("log-facility", "LOG_LOCAL6", 0),
		CFG_STR("firewall-program", 0, 0),
		CFG_INT("firewall-soft-timeout", 10, 0),
		CFG_INT("firewall-hard-timeout", 3, 0),
		CFG_INT("port", 1717, 0),
		CFG_STR("oasis-host", 0, 0),
		CFG_INT("oasis-port", 1718, 0),
		CFG_STR("probe-library", "libprobe_ping.so", 0),
		CFG_STR("probe-interface", "eth0", 0),
		CFG_STR("network", 0, 0),
		CFG_INT("promiscuous", 0, 0),
		CFG_END()
	};

	while((c = getopt_long(argc, argv, "fr:dhV", longopts, 0)) != EOF)
	{
		switch(c) {
			case 'f':
				foreground = 1;
				break;
			case 'r':
				rcfile = optarg;
				break;
			case 'd':
				use_debug = 1;
				break;
			case 'h':
				printf(COPYLINE);
				printf(SYNTAX);
				return 0;
				break;
			case 'V':
				printf("fwcd " VERSION "\n");
				return 0;
				break;
			case '?':
			default:
				return 1;
		}
	}

	printf(COPYLINE);

	fprintf(stderr, "parsing configuration file %s...\n", rcfile);
	cfg_set_error_function(logmsg_cfg);
	cfg = cfg_init(opts, 0);
	cfg_parse(cfg, rcfile);

	if(geteuid() != 0) {
		fprintf(stderr, "This program must be run with root privileges\n");
		return 2;
	}

	install_signal_handlers();

	program_name = "fwcd"; /* used in logmsg() in common.c */

	if(cfg_getstr(cfg, "network") == 0) {
		fprintf(stderr, "no network specified in config file -- exiting");
		return 4;
	}

	if(foreground) {
		you_can_dance_now();
	} else {
		pid = fork();
		if(pid == 0) {
			/* Make sure we have stdin/stdout/stderr
			 */

			close(0);
			close(1);
			close(2);
			open("/dev/null", 0);
			open("/dev/null", 0);
			open("/dev/null", 0);

			/* Initialize logging
			 */
			openlog(program_name, LOG_PID,
					log_facility(cfg_getstr(cfg, "log-facility")));
			use_syslog = 1;

			/* ...and go
			 */
			you_can_dance_now();
		} else if(pid == -1) {
			fprintf(stderr, "Failed to fork(): %s", strerror(errno));
			return 1;
		}
	}

	return 0;
}
