/* Open.Net Authentication Daemon
 * SSL client/server functions
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "common.h"
#include "fwssl.h"

static int passwd_cb(char *buf, int size, int rwflag, void *password)
{
	strncpy(buf, (char *)(password), size);
	buf[size - 1] = '\0';
	return strlen(buf);
}

void logmsg_ssl(int fac)
{
	long e;
	while((e = ERR_get_error()) != 0)
		logmsg(LOG_ERR, "%s", ERR_error_string(e, 0));
}

SSL_CTX *init_ctx(const char *certfile,
				  const char *keyfile,
				  const char *valid_cert_file,
				  int client)
{
	SSL_CTX *ctx;

	SSL_load_error_strings();
	SSL_library_init();

	ctx = SSL_CTX_new(client? SSLv23_client_method() : SSLv23_server_method());
	if(!ctx) {
		logmsg_ssl(LOG_ERR);
		return 0;
	}

	SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, 0);

	if(!client) {
		/* In server mode, when requesting a client certificate, the
		 * server must send the list of CAs of which it will accept client
		 * certificates. This list is not influenced by the contents of
		 * CAfile or CApath and must explicitly be set using the
		 * SSL_CTX_set_client_CA_list(3) family of functions.
		 */
/*		SSL_CTX_set_client_CA_list(ctx, );*/
	}
	
	if(valid_cert_file)
		SSL_CTX_load_verify_locations(ctx, valid_cert_file, 0);

	if(certfile && keyfile) {
		logmsg(LOG_INFO, "Using certificate from %s, and private key from %s",
			   certfile, keyfile);

		SSL_CTX_set_default_passwd_cb(ctx, passwd_cb);
        SSL_CTX_set_default_passwd_cb_userdata(ctx, (void *)"");

		if(SSL_CTX_use_certificate_file(ctx, certfile, SSL_FILETYPE_PEM) <= 0){
			logmsg_ssl(LOG_ERR);
			return 0;
		}

		if(SSL_CTX_use_PrivateKey_file(ctx, keyfile, SSL_FILETYPE_PEM) <= 0) {
			logmsg_ssl(LOG_ERR);
			return 0;
		}

		if(!SSL_CTX_check_private_key(ctx)) {
			logmsg(LOG_ERR,
				   "Private key does not match the certificate public key");
			return 0;
		}
	} else
		logmsg(LOG_INFO, "No certificates configured -- proceeding anyway");

	return ctx;
}

int check_certificate(SSL *ssl)
{
	X509 *cert;
	int ret = 1;

	logmsg(LOG_INFO, "Using %s", SSL_get_cipher(ssl));

	/* Get servers's certificate */
	cert = SSL_get_peer_certificate(ssl);
	if(cert != NULL) {
		char *str;

		logmsg(LOG_INFO, "peer certificate:");

		str = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
		logmsg(LOG_INFO, "subject: %s", str);
		free(str);

		str = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);
		logmsg(LOG_INFO, "issuer: %s", str);
		free(str);

		if(SSL_get_verify_result(ssl) != X509_V_OK) {
			logmsg_ssl(LOG_ERR);
			logmsg(LOG_ERR, "failed to verify peer certificate!");
		} else
			ret = 0;

		X509_free(cert);
	} else {
		logmsg(LOG_INFO, "Peer does not have a certificate -- aborting");
		ret = 2;
	}

	return ret;
}

int init_ssl_daemon(SSL_CTX *ctx, int port, int (*callback)(SSL *))
{
	int fd;
	struct sockaddr_in addr;
	int cfd;
	struct sockaddr_in client_addr;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	SSL *ssl;
	int on = 1;

	/* create a TCP socket */
	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(fd == -1) {
		logmsg(LOG_ERR, "failed to create socket: %s", strerror(errno));
		return 1;
	}

	/* enable local address reuse */
	if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1)
		logmsg(LOG_DEBUG, "setsockopt(SO_REUSEADDR) failed, ignoring");

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;
	if(bind(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) == -1) {
		logmsg(LOG_ERR, "failed to bind socket to port %d: %s", port,
			   strerror(errno));
		return 1;
	}

	if(listen(fd, 1) == -1) {
		logmsg(LOG_ERR, "listen: %s", strerror(errno));
		return 1;
	}

	logmsg(LOG_INFO, "listening for connections on port %d...", port);

	cfd = accept(fd, (struct sockaddr *)&client_addr, &addrlen);
	logmsg(LOG_INFO, "got connection from %s:%d",
		   inet_ntoa(client_addr.sin_addr), ntohs(client_addr.sin_port));

	/* close the original socket, we don't need it anymore */
	close(fd);

	ssl = SSL_new(ctx);
	if(!ssl) {
		logmsg_ssl(LOG_ERR);
		return 1;
	}

	/* assign the network socket to the ssl object */
	if(SSL_set_fd(ssl, cfd) == 0) {
		logmsg_ssl(LOG_ERR);
		return 1;
	}

	if(SSL_accept(ssl) != 1) {
		logmsg_ssl(LOG_ERR);
		return 1;
	}

	if(check_certificate(ssl) != 0) {
		SSL_shutdown(ssl);
		close(cfd);
		return 2;
	}

	while(callback(ssl) != -2)
		/* loop */ ;
	SSL_shutdown(ssl);
	close(cfd);

	return 0;
}

SSL *init_ssl_client(SSL_CTX *ctx, const char *host, int port)
{
	int fd;
	struct sockaddr_in addr;
	SSL *ssl;
	struct hostent *he;

	if(!host) {
		logmsg(LOG_ERR, "Null host in init_ssl_client");
		return 0;
	}

	/* create a TCP socket */
	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(fd == -1) {
		logmsg(LOG_ERR, "failed to create socket: %s", strerror(errno));
		return 0;
	}

	/* look up the hostname */
	he = gethostbyname(host);
	if(!he) {
		logmsg(LOG_ERR, "%s: %s", host, hstrerror(h_errno));
		close(fd);
		return 0;
	}
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	memcpy((void *)&addr.sin_addr, he->h_addr_list[0], he->h_length);

	logmsg(LOG_INFO, "Connecting to server at %s:%d...",
		   inet_ntoa(addr.sin_addr), port);

	if(connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) == -1)
	{
		logmsg(LOG_ERR, "%s: %s", host, strerror(errno));
		close(fd);
		return 0;
	}

	ssl = SSL_new(ctx);
	if(!ssl) {
		logmsg_ssl(LOG_ERR);
		close(fd);
		return 0;
	}

	/* assign the network socket to the ssl object */
	if(SSL_set_fd(ssl, fd) == 0) {
		logmsg_ssl(LOG_ERR);
		close(fd);
		return 0;
	}

	/* initiate the SSL handshake */
	if(SSL_connect(ssl) != 1) {
		logmsg_ssl(LOG_ERR);
		close(fd);
		return 0;
	}

	if(check_certificate(ssl) != 0) {
		SSL_shutdown(ssl);
		close(fd);
		return 0;
	}

	return ssl;
}

/* returns: 0 (success), 1 (reconnect and retry), 2 (give up)
 */
int check_ssl_io(int ret, SSL *ssl, int expected)
{
	if(ret > 0) {
		/* success */
		if(ret != expected) {
			logmsg(LOG_ERR, "short read/write: %d != %d", ret, expected);
			return 2;
		}
		return 0;
	} else if(ret == 0) {
		/* The SSL connection has been closed */
		logmsg(LOG_WARNING, "SSL connection closed by peer");
		return 1;
	} else {
		int e = SSL_get_error(ssl, ret);
		if(e == SSL_ERROR_SSL || e == SSL_ERROR_SYSCALL)
			logmsg_ssl(LOG_ERR);
		else {
			logmsg(LOG_ERR, "can't handle this SSL error: %d", e);
			/* am I lazy? */
		}
		return 2;
	}
}


req_status_t fw_open_local(char *firewall,
						   struct in_addr *ip,
						   unsigned char mac[6])
{
	int fwstatus;
	char *ipstr = inet_ntoa(*ip);
	char * const argv[] = {firewall, "open", ipstr, strmac(mac), 0};
	req_status_t status = REQ_STATUS_OK;

	fwstatus = spawn(argv);
	if(!WIFEXITED(fwstatus)) {
		/* child did not exit normally */
		logmsg(LOG_ERR, "%s(open) exited abnormally", firewall);
		status = REQ_STATUS_FWFAILED;
	} else {
		if(WEXITSTATUS(fwstatus) != 0) {
			logmsg(LOG_ERR, "%s(open) exited with error code %d",
				   firewall, WEXITSTATUS(fwstatus));
			status = REQ_STATUS_FWFAILED;
		} else {
			/* firewall control program succeeded
			 */
			status = REQ_STATUS_OK;
		}
	}
	return status;
}

req_status_t fw_close_local(char *firewall,
							struct in_addr *ip,
							unsigned char mac[6])
{
	int fwstatus;
	char *ipstr = inet_ntoa(*ip);
	char * const argv[] = {firewall, "close", ipstr, strmac(mac), 0};
	req_status_t status = REQ_STATUS_OK;

	fwstatus = spawn(argv);
	if(!WIFEXITED(fwstatus)) {
		/* child did not exit normally */
		logmsg(LOG_ERR, "%s(close) exited abnormally", firewall);
		status = REQ_STATUS_FWFAILED;
	} else {
		if(WEXITSTATUS(fwstatus) != 0) {
			logmsg(LOG_ERR, "%s(close) exited with error code %d",
				   firewall, WEXITSTATUS(fwstatus));
			status = REQ_STATUS_FWFAILED;
		} else {
			status = REQ_STATUS_OK;
		}
	}
	return status;
}

req_status_t fw_reset_local(char *firewall)
{
	char * const argv[] = {firewall, "reset", 0};
	spawn(argv);
	/* return value is ignored
	 */
	return REQ_STATUS_OK;
}
