/* $Id: list.c,v 1.7 2002/02/21 17:42:17 mhe Exp $
 *
 * list.c -- generic linked list
 *
 * Yet Another FTP Client
 * Copyright (C) 1998-2001, Martin Hedenfalk <mhe@stacken.kth.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. See the GPL for more details.
 */

#include "common.h"
#include "list.h"
#include "xmalloc.h"

list_t *list_new(list_delfunc_t delfunc)
{
	list_t *lp = (list_t *)xmalloc(sizeof(list_t));

	lp->delfunc = delfunc;
	lp->size = 0;
	lp->first = 0;

	return lp;
}

void list_clear(list_t *lp)
{
	if(!lp)
		return;
	while(lp->first)
		list_remove(lp, lp->first);
}

void list_destroy(list_t *lp)
{
	list_clear(lp);
	xfree(lp);
}

/* Search for an item in the list. The comparison function should
 * return 0 if an item from the list matches with the supplied
 * argument. The comparison function is called with the data from an
 * item in the list as first argument and the supplied argument ARG as
 * second argument. Returns NULL if the item was not found, otherwise
 * the item is returned.
 */
listitem_t *list_search(list_t *lp, list_cmpfunc_t cmpfunc, const void *arg)
{
	listitem_t *lip;

	if(!lp)
		return 0;

	lip = lp->first;
	while(lip) {
		if(cmpfunc(lip->data, arg) == 0)
			return lip;
		lip = lip->next;
	}
	return 0;
}

listitem_t *list_detach(list_t *lp, listitem_t *li)
{
	if(!lp || !lp->first)
		return 0;

	if(li->prev)
		li->prev->next = li->next;
	else
		lp->first = li->next;
	if(li->next)
		li->next->prev = li->prev;
	else
		lp->last = li->prev;
	lp->size--;

	return li;
}

/* Remove an item from the list and deallocate the memory for that
 * item if a destructor function is set
 */
void list_remove(list_t *lp, listitem_t *lip)
{
	if(!lp || !lp->first)
		return;

	if(lip->prev)
		lip->prev->next = lip->next;
	else
		lp->first = lip->next;
	if(lip->next)
		lip->next->prev = lip->prev;
	else
		lp->last = lip->prev;
	if(lp->delfunc)
		lp->delfunc(lip->data);
	xfree(lip);
	lp->size--;
}

/* Create and insert a new item with data into the end of the list
 */
listitem_t *list_insert(list_t *lp, void *data)
{
	listitem_t *li = (listitem_t *)xmalloc(sizeof(listitem_t));

	if(!lp)
		return 0;

	li->data = data;
	li->next = 0;
	li->prev = 0;

	if(lp->first == 0) {
		lp->last = lp->first = li;
	} else {
		li->prev = lp->last;
		li->prev->next = li;
		lp->last = li;
	}
	lp->size++;
	return li;
}

int list_size(list_t *lp)
{
	if(!lp)
		return 0;
	return lp->size;
}

static void list_swap(list_t *lp, listitem_t *a, listitem_t *b)
{
	void *tmp;

	if(!a || !b || a->data == b->data)
		return;

	tmp = a->data;
	a->data = b->data;
	b->data = tmp;
}

/* simple bubble-sort */
void list_sort(list_t *lp, list_cmpfunc_t cmp, int reverse)
{
	listitem_t *li, *last = 0;
	int swapped = 1;
	int c;

	while(swapped) {
		swapped = 0;
		for(li=lp->first; li && li->next; li=li->next) {
			if(li->next == last) {
				last = li;
				break;
			}
			c = cmp(li->data, li->next->data);
			if(reverse ? c < 0  : c > 0) {
				list_swap(lp, li, li->next);
				swapped = 1;
			}
		}
	}
}

void list_movetofront(list_t *lp, listitem_t *li)
{
	li = list_detach(lp, li);
	if(!li)
		return;

	li->next = 0;
	li->prev = 0;

	if(lp->first == 0) {
		lp->last = lp->first = li;
	} else {
		li->next = lp->first;
		li->next->prev = li;
		lp->first = li;
	}
	lp->size++;
}

void list_movetoback(list_t *lp, listitem_t *li)
{
	li = list_detach(lp, li);
	if(!li)
		return;

	li->next = 0;
	li->prev = 0;

	if(lp->first == 0) {
		lp->last = lp->first = li;
	} else {
		li->prev = lp->last;
		li->prev->next = li;
		lp->last = li;
	}
	lp->size++;
}
