/* Open.Net Authentication Daemon
 * Initialization
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: main.c,v 1.24 2002/04/25 12:03:32 mhe Exp $
 */

#include "common.h"

cfg_t *cfg = 0;
int num_domain = 0;

#define _GNU_SOURCE
#include <getopt.h>

#define COPYLINE PACKAGE" version "VERSION"\n" \
"Copyright (c) 2002, Royal Institute of Technology. All rights reserved.\n\n"

#define SYNTAX \
"Usage: "PACKAGE" [options]\n" \
"Options:\n" \
"  -f, --fg            run in foreground (don't fork, send messages to terminal)\n" \
"  -r, --rcfile=FILE   read configuration from FILE\n" \
"  -d, --debug         include debug messages in syslog\n" \
"  -F, --force         continue even if socket already exists\n" \
"  -h, --help          show this help and exit successfully\n" \
"  -V, --version       show version information and exit successfully\n" \
"\n" \
"Report bugs to <open-devel@stockholmopen.net>\n" \
"\n"

void cleanup(void)
{
	unlink(cfg_getstr(cfg, "socket-path"));
	reset_firewalls();
	monitor_shutdown();
#ifdef HAVE_LIBDMALLOC
	dmalloc_shutdown();
#endif
}

void sigterm_handler(int signo)
{
	logmsg(LOG_INFO, "Caught signal %d, shutting down...", signo);
	cleanup();

	/* Now reraise the signal.  We reactivate the signal's default
	 * handling, which is to terminate the process. We could just
	 * call `exit' or `abort', but reraising the signal sets the
	 * return status from the process correctly.
	 */
	signal(signo, SIG_DFL);
	raise(signo);
}

void sigusr1_handler(int signo)
{
	use_debug = !use_debug;
	logmsg(LOG_INFO, "Debugging is now turned %s", use_debug ? "on" : "off");
	signal(SIGUSR1, sigusr1_handler);
}

/* Setup signal handlers
 */
void install_signal_handlers(void)
{
	struct sigaction action;
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, SIGINT);
	sigaddset(&mask, SIGQUIT);
	action.sa_handler = sigterm_handler;
	action.sa_mask = mask;
	action.sa_flags = 0;
	sigaction(SIGTERM, &action, NULL);

	sigemptyset(&mask);
	sigaddset(&mask, SIGTERM);
	sigaddset(&mask, SIGQUIT);
	action.sa_handler = sigterm_handler;
	action.sa_mask = mask;
	action.sa_flags = 0;
	sigaction(SIGINT, &action, NULL);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGUSR1, sigusr1_handler);
}

void sighup_handler(int signo)
{
	logmsg(LOG_INFO, "========================= GOT SIGHUP");
	signal(SIGHUP, SIG_IGN); /* ignore signals */
	signal(SIGTSTP, SIG_IGN);

	/* Make sure we have stdin/stdout/stderr
	 */
	close(0);
	close(1);
	close(2);
	open("/dev/null", 0);
	open("/dev/null", 0);
	open("/dev/null", 0);

	/* Initialize logging
	 */
	openlog(program_name, LOG_PID,
			log_facility(cfg_getstr(cfg, "log-facility")));
	use_syslog = 1;
}

int main(int argc, char **argv)
{
	int pid;
	struct option longopts[] = {
		{"fg", no_argument, 0, 'f'},
		{"rcfile", required_argument, 0, 'r'},
		{"debug", no_argument, 0, 'd'},
		{"force", no_argument, 0, 'F'},
		{"quiet", no_argument, 0, 'q'},
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'V'},
		{0, 0, 0, 0}
	};
	char *sockfile;
	int c;
	int foreground = 0; /* run in foreground (don't fork) */
	char *rcfile = SYSCONFDIR "/oasis.conf";
	int force = 0;
	int quiet = 0;
	cfgval_t *cval = 0;
	cfgopt_t domain_opts[] = {
		CFG_INT("max-users", -1, 0),
		CFG_INT("max-same-users", 1, 0),
		CFG_STR("firewall-program", 0, 0),
		CFG_STR("probe-library", "libprobe_ping.so", 0),
		CFG_STR("probe-interface", "eth0", 0),
		CFG_STR("fwcd-host", 0, 0),
		CFG_INT("fwcd-port", 1717, 0),
		CFG_INT("port", 1718, 0),
		CFG_STR("certificate-file", SYSCONFDIR "/oasis.cert", 0),
		CFG_STR("key-file", SYSCONFDIR "/oasis.key", 0),
		CFG_STR("fwcd-certificate-file", SYSCONFDIR "/fwcd.cert", 0),
		CFG_STR("pam-service", "login", 0),
		CFG_END()
	};
	cfgopt_t opts[] = {
		CFG_INT("probe-interval", 10, 0),
		CFG_INT("probe-max-users", 100, 0),
		CFG_INT("probe-timeout", 8, 0),
		CFG_INT("max-missed-probes", 3, 0),
		CFG_INT("backlog", 50, 0),
		CFG_INT("firewall-soft-timeout", 15, 0),
		CFG_INT("firewall-hard-timeout", 3, 0),
		CFG_STR("log-facility", "LOG_LOCAL6", 0),
		CFG_STR("socket-path", "/tmp/oasis", 0),
		CFG_STR("socket-mode", "0600", 0),
		CFG_STR("socket-owner", "nobody.nobody", 0),
		CFG_SEC("domain", domain_opts, CFGF_MULTI | CFGF_TITLE),
		CFG_END()
	};

	while((c = getopt_long(argc, argv, "fr:dFqhV", longopts, 0)) != EOF)
	{
		switch(c) {
			case 'f':
				foreground = 1;
				break;
			case 'r':
				rcfile = optarg;
				break;
			case 'd':
				use_debug = 1;
				break;
			case 'F':
				force = 1;
				break;
			case 'q':
				quiet = 1;
				break;
			case 'h':
				printf(COPYLINE);
				printf(SYNTAX);
				return 0;
				break;
			case 'V':
				printf(PACKAGE " " VERSION "\n");
				return 0;
				break;
			case '?':
			default:
				return 1;
		}
	}

	if(!quiet)
		printf(COPYLINE);

	if(geteuid() != 0) {
		fprintf(stderr,
				"This program must be run with root privileges -- exiting\n");
		exit(2);
	}

	if(!quiet)
		fprintf(stderr, "parsing configuration file %s...\n", rcfile);
	cfg_set_error_function(logmsg_cfg);
	cfg = cfg_init(opts, 0);
	cfg_parse(cfg, rcfile);

	sockfile = cfg_getstr(cfg, "socket-path");

	if(!force && access(sockfile, F_OK) == 0) {
		fprintf(stderr, "socket %s already exists -- exiting\n", sockfile);
		exit(1);
	}

	/* perform some sanity checks on the config values
	 */

	if((c = cfg_getint(cfg, "backlog")) <= 0) {
		fprintf(stderr, "backlog value too small (%d) -- exiting\n", c);
		exit(6);
	}
	cval = cfg_getval(cfg, "domain");
	if(!cval || cval->a.size <= 0) {
		fprintf(stderr, "no domains defined -- exiting\n");
		exit(6);
	}
	num_domain = cval->a.size;
	install_signal_handlers();

	program_name = "oasis";
	oasis_start = time(0);

	if(foreground) {
		signal(SIGHUP, sighup_handler);
		monitor_init();
		request_loop();
	} else {
		pid = fork();
		if(pid == 0) {
			/* Make sure we have stdin/stdout/stderr
			 */
			close(0);
			close(1);
			close(2);
			open("/dev/null", 0);
			open("/dev/null", 0);
			open("/dev/null", 0);

			/* Initialize logging
			 */
			openlog(PACKAGE, LOG_PID,
					log_facility(cfg_getstr(cfg, "log-facility")));
			use_syslog = 1;

			/* ...and go
			 */
			monitor_init();
			request_loop();

		} else if(pid == -1) {
			logmsg(LOG_ERR, "Failed to fork(): %s", strerror(errno));
			return 1;
		}
	}

	return 0;
}
