/* Open.Net Authentication Daemon
 * Monitoring functions
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: monitor.c,v 1.38 2002/04/25 11:07:53 mhe Exp $
 */

#include "common.h"
#include "list.h"
#include "xmalloc.h"

#include "fwssl.h"

#include <dlfcn.h>

static pthread_t monitor_thread;
static pthread_t fwcd_thread;
pthread_mutex_t modify_list_mutex = PTHREAD_MUTEX_INITIALIZER;
domain_t *domains = 0;
list_t *users = 0;
list_t *logout_users = 0;
pthread_mutex_t logout_users_mutex = PTHREAD_MUTEX_INITIALIZER;

#ifndef timersub
# define timersub(a, b, result)                                               \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                             \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                          \
    if ((result)->tv_usec < 0) {                                              \
      --(result)->tv_sec;                                                     \
      (result)->tv_usec += 1000000;                                           \
    }                                                                         \
  } while (0)
#endif

/* destructor for the user list
 */
void userdata_destroy(userdata_t *ud)
{
	if(ud) {
		xfree(ud->username);
		ud->username = "USING free()'d MEMORY!";
		ud->domain = (domain_t *)0xDEADBEEF;
		xfree(ud);
	}
}

static domain_t *find_domain(char *domain)
{
	int i;

	for(i = 0; i < num_domain; i++) {
		if(strcmp(domain, domains[i].name) == 0)
			return &domains[i];
	}
	return 0;
}

/* create a new userdata structure
 */
userdata_t *userdata_create(char *username, char *domain,
							char *password, struct in_addr *ip,
							unsigned char mac[])
{
	userdata_t *ud;

	ud = xmalloc(sizeof(userdata_t));
	memcpy(ud->mac, mac, MAX_MAC_LEN);
	ud->ip = *ip;
	ud->username = xstrdup(username);
	ud->domain = find_domain(domain); /* might return NULL (no such domain) */
	ud->mstatus = MS_OK;
	ud->missed_probes = 0;
	ud->failed_logouts = 0;
	ud->probe_id = 0;
	timerclear(&ud->tv);
	ud->probe_data = (void *)0xDEADBEEF;
	pthread_mutex_init(&ud->mutex, 0);
	timerclear(&ud->logintime);

	return ud;
}

/* search functions for the user list
 */
static int user_search_mac(userdata_t *ud, userdata_t *arg)
{
	return memcmp(ud->mac, arg->mac, sizeof(arg->mac));
}

static int user_search_mac_domain(userdata_t *ud, userdata_t *arg)
{
	int r;

	r = memcmp(ud->mac, arg->mac, sizeof(arg->mac));
	if(r == 0) {
		/* This allows a user to be logged in with the *SAME* MAC
		 * address in different domains!
		 */
		r = strcmp(ud->domain->name, arg->domain->name);
	}
	return r;
}

static void probe_missed(userdata_t *ud)
{
	/* user should be in the MS_UPDATING state when entering this
	 * function.
	 */

	/* Here we don't want to be cancelled, otherwise we might
	 * end up with a not fully logged out user.
	 */
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, 0);

	if(++ud->missed_probes >= cfg_getint(cfg, "max-missed-probes"))
	{
		req_status_t status;
		status = logout(ud);
		if(status != REQ_STATUS_OK) {
			/* We failed to logout the user! What now? We reset the
			 * user to the MS_OK state and probes another time
			 * (without resetting the missed_probes counter)
			 */
			/* Update a counter of failed logouts, and force removal
			 * from the list after three failed logouts. This could
			 * happen if oasis and the firewall is out of sync (eg,
			 * user logged in according to oasis, but not according to
			 * the firewall).
			 */
			if(++ud->failed_logouts >= 3) {
				logmsg(LOG_ERR,
					   "logout failed (%s) for %s@%s, forcing removal",
					   strstatus(status), ud->username, ud->domain->name);
				status = monitor_remove(ud);
				if(status != REQ_STATUS_OK) {
					/* Oops, wasn't even able to remove the user from
					 * the list, this should not happen
					 */
					logmsg(LOG_ERR, "Can't remove %s@%s: %s",
						   ud->username, ud->domain->name, strstatus(status));
					set_mstatus(ud, MS_OK);
				}
			} else {
				logmsg(LOG_ERR,
					   "logout failed (%s) for %s@%s, resetting user status",
					   strstatus(status), ud->username, ud->domain->name);
				set_mstatus(ud, MS_OK);
			}
		}
	} else {
		logmsg(LOG_DEBUG, "%s@%s missed a probe",
			   ud->username, ud->domain->name);
		set_mstatus(ud, MS_OK);
	}

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
}

static void *probe_thread_main(userdata_t *ud)
{
	int (*func)(userdata_t *, int);
	int ret;

	set_mstatus(ud, MS_PROBING);

	/* enable immediate thread cancellation  */
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);
	/* Call the dynamically loaded probe function (ping) */
	func = ud->domain->probe_user_func;
	ret = (*func)(ud, 1);
	/* disable thread cancellation */
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, 0);

	if(ret == 0) {
		/* positive response, we got a reply, reset the counter */
		ud->missed_probes = 0;
		set_mstatus(ud, MS_OK);
	} else {
		/* user missed the probe */
		set_mstatus(ud, MS_UPDATING);
		probe_missed(ud);
	}

	if(ud->domain->probe_cleanup_func)
		(*ud->domain->probe_cleanup_func)(ud);

	return 0;
}

/* Checks if the user can login, ie if this user is already logged in
 * or if some maximum has been reached. This is not authentication.
 */
req_status_t monitor_check_user(userdata_t *ud)
{
	req_status_t status = REQ_STATUS_OK;
	int maxusers;

	/* we need to sync the user information in the access server and
	 * the firewall control daemon
	 */

	pthread_mutex_lock(&modify_list_mutex);

	/* Check if the user already exists in the specified domain
	 */
	maxusers = cfg_stgetint(cfg, "domain", ud->domain->name, "max-users");
	if(maxusers != -1 && ud->domain->nusers >= maxusers)
		status = REQ_STATUS_REJFULL;
	else if(list_search(users, (list_cmpfunc_t)user_search_mac_domain,
						(void *)ud) != 0)
		status = REQ_STATUS_USEREXIST;
	else {
		listitem_t *li;
		int maxlogins = cfg_stgetint(cfg, "domain", ud->domain->name,
									 "max-same-users");
		int numlogins = 0;

		/* Check if the user is already connected from another
		 * machine, and if the number of logins is within the
		 * configured limit. This searches the user list for an entry
		 * with the same _user@domain_ (not MAC address)
		 */
		if(maxlogins != -1) {
			for(li = users->first; li; li = li->next) {
				if(strcmp(ud->username,
						  ((userdata_t *)li->data)->username) == 0 &&
				   ud->domain == ((userdata_t *)li->data)->domain)
				{
					if(++numlogins >= maxlogins) {
						status = REQ_STATUS_USEREXIST;
						break;
					}
				}
			}
		}
	}
	pthread_mutex_unlock(&modify_list_mutex);

	return status;
}

/* insert a user into the monitoring list
 */
req_status_t monitor_insert(userdata_t *ud)
{
	pthread_mutex_lock(&modify_list_mutex);
	list_insert(users, ud);
	ud->domain->nusers++;
	pthread_mutex_unlock(&modify_list_mutex);

	return REQ_STATUS_OK;
}

/* remove a user from the monitoring list
 */
req_status_t monitor_remove(userdata_t *ud)
{
	req_status_t status = REQ_STATUS_FAIL;
	listitem_t *li;

	pthread_mutex_lock(&modify_list_mutex);

	li = list_search(users, (list_cmpfunc_t)user_search_mac, ud);
	if(li) {
		ud->domain->nusers--;
		if(ud->domain->fwtype == FW_LOCAL) {
			/* Here we detach the user rather than removing it. The
			 * difference is that we don't free the memory used by the
			 * user at this point. Instead, it's free'd by the probe
			 * checking in monitor_thread_main().
			 */
			list_detach(users, li);
			set_mstatus(ud, MS_DETACHED);
		} else {
			/* For users controled by a remote firewall, we can remove
			 * the user and free the memory directly
			 */
			list_remove(users, li);
		}
		status = REQ_STATUS_OK;
	} else
		status = REQ_STATUS_NOUSER;

	pthread_mutex_unlock(&modify_list_mutex);
	return status;
}

void monitor_shutdown(void)
{
	int i;

	logmsg(LOG_INFO, "shutting down monitor");
	if(pthread_cancel(monitor_thread) != 0)
		logmsg(LOG_ERR, "failed to cancel monitoring thread");
	/* wait for the monitor thread to exit */
	pthread_join(monitor_thread, 0);

	if(pthread_cancel(fwcd_thread) != 0)
		logmsg(LOG_ERR, "failed to cancel fwcd thread");
	/* wait for the monitor thread to exit */
	pthread_join(fwcd_thread, 0);

	/* deallocate the domains */
	for(i = 0; i < num_domain; i++) {
		if(domains[i].name) {
			logmsg(LOG_DEBUG, "deallocating domain %s", domains[i].name);

			xfree(domains[i].name);
			domains[i].name = 0;
			if(domains[i].probelib_handle)
				dlclose(domains[i].probelib_handle);
			else if(domains[i].sslout)
				SSL_shutdown(domains[i].sslout);
			domains[i].probelib_handle = 0;
			domains[i].sslout = 0;
		}
	}

/*	list_destroy(users);*/
}

static void monitor_thread_cleanup(void *arg)
{
	listitem_t *li;

	logmsg(LOG_DEBUG, "cancelling monitor thread");

	/* we must cancel any probe threads runing before we can
	 * destroy the domain lists
	 */
	for(li = users->first; li; li = li->next) {
		userdata_t *ud = (userdata_t *)li->data;
		if(ud->probe_id && get_mstatus(ud) != MS_OK &&
		   pthread_cancel(ud->probe_id))
		{
			logmsg(LOG_WARNING, "Failed to cancel probe thread %ld for %s@%s",
				   ud->probe_id, ud->username, ud->domain->name);
		} else {
			if(ud->domain->probe_cleanup_func)
				(*ud->domain->probe_cleanup_func)(ud);
		}
	}
}

static void cancel_probe(userdata_t *ud)
{
	logmsg(LOG_DEBUG, "cancelling probe thread %d for %s@%s",
		   ud->probe_id, ud->username, ud->domain->name);
	if(pthread_cancel(ud->probe_id) == 0) {
		pthread_attr_t thread_attr;
		if(ud->domain->probe_cleanup_func)
			(*ud->domain->probe_cleanup_func)(ud);
		set_mstatus(ud, MS_UPDATING);
		pthread_attr_init(&thread_attr);
		pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&ud->probe_id, 0,
					   (void *(*)(void *))probe_missed,
					   ud);
		pthread_attr_destroy(&thread_attr);
	} else
		logmsg(LOG_ERR, "unable to cancel runaway probe thread %ld: %s",
			   ud->probe_id, strerror(errno));
}

static void wait_until(struct timeval *tv)
{
	struct timeval now, res;

	gettimeofday(&now, 0);
	if(timercmp(&now, tv, >=))
		return;
	timersub(tv, &now, &res);
	select(0, 0, 0, 0, &res);
}

/* The monitoring thread is responsible of probing the users to see if
 * they are still connected. probe-max-users are probed at the same
 * time each probe-interval seconds. Each user is allowed to miss at
 * the most max-missed-probes probes without being logged out.
 */
static void *monitor_thread_main(void *arg)
{
	list_t *prio = 0;
	struct timeval next_probe;
	int maxprobes;
	pthread_attr_t thread_attr;

	block_term_sigs();

	/* enable immediate thread cancellation
	 */
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);
	pthread_cleanup_push(monitor_thread_cleanup, (void *)0);

	logmsg(LOG_INFO, "monitoring thread starting");

	/* Create a list used a priority queue. Users being probed are
	 * linked to this list when a probe thread has started for that
	 * user and are checked below.
	 */
	prio = list_new(0);

	/* Initialize a thread attribute to be used when creating probe
	 * threads below
	 */
	pthread_attr_init(&thread_attr);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);

	while(1) {
		int nprobes;
		int i;
		listitem_t *li;

		/* Decide the next action, either to cancel a probe or
		 * continue on with the next probe session. Check the prio
		 * queue to see if there are any probes that times out before
		 * the time for the next probe session, otherwise wait and go
		 * on with the next probe session.
		 */

		gettimeofday(&next_probe, 0);
		next_probe.tv_sec += cfg_getint(cfg, "probe-interval");

		while(prio->first &&
			  timercmp(&((userdata_t*)prio->first->data)->tv, &next_probe, <=))
		{
			userdata_t *ud = (userdata_t *)prio->first->data;
			mstatus_t mstatus;

			/* We must process an old probe before continuing with the
			 * next probe session.
			 */

			wait_until(&ud->tv);

			mstatus = get_mstatus(ud);
			switch(mstatus) {
				case MS_PROBING:
					cancel_probe(ud);
					list_remove(prio, prio->first);
					break;
				case MS_OK:
					/* We got a successful reply and the thread should
					 * have exited by itself. Remove it from the prio
					 * queue.
					 */
					ud->probe_id = 0;
					list_remove(prio, prio->first);
					break;
				case MS_DETACHED:
					/* The user has already been logged out. Now we can
					 * free the memory used by the user.
					 */
					list_remove(prio, prio->first);
					userdata_destroy(ud);
					break;
				case MS_UPDATING:
				case MS_LOGOUT:
					/* The probe is (about to) log out a user, let it continue
					 */
					logmsg(LOG_DEBUG,
						   "skipping cancellation of probe for %s@%s (%s)",
						   ud->username, ud->domain->name,
						   strmstatus(mstatus));
					/* Update the timer for this probe and place it
					 * last in the prio list. When the user is finally
					 * logged out, it is in the MS_DETACHED state and
					 * we must free the memory.
					 */
					gettimeofday(&ud->tv, 0);
					ud->tv.tv_sec += cfg_getint(cfg, "probe-timeout");
					list_movetoback(prio, prio->first);
					break;
			}
		}
		wait_until(&next_probe);

		/* Ok, go on with another probe session. But first we have to
		 * remove finished probes from the prio list.
		 */
		for(li = prio->first; li; /* nothing */) {
			if(get_mstatus((userdata_t *)li->data) == MS_OK) {
				listitem_t *xli = li->next;
				((userdata_t *)li->data)->probe_id = 0;
				list_remove(prio, li);
				li = xli;
			} else
				li = li->next;
		}

		/* Logout queued users in the logout_users list.
		 */
		pthread_mutex_lock(&logout_users_mutex);
		for(li = logout_users->first; li; li = li->next) {
			userdata_t *ud = (userdata_t *)li->data;
			req_status_t status;

			logmsg(LOG_DEBUG, "requested logout of %s@%s",
				   ud->username, ud->domain->name);
			status = logout(ud);

			if(status != REQ_STATUS_OK) {
				logmsg(LOG_ERR,
					   "logout failed (%s) for %s@%s, forcing removal",
					   strstatus(status), ud->username, ud->domain->name);
				status = monitor_remove(ud);
				if(status != REQ_STATUS_OK) {
					/* Oops, wasn't even able to remove the user from
					 * the list, this should not happen
					 */
					logmsg(LOG_ERR, "Can't remove %s@%s: %s",
						   ud->username, ud->domain->name, strstatus(status));
					set_mstatus(ud, MS_OK);
				}
			}
		}
		list_clear(logout_users);
		pthread_mutex_unlock(&logout_users_mutex);

		maxprobes = cfg_getint(cfg, "probe-max-users");

		/* Make sure we unlock the mutex if this thread is cancelled.
		 * In what other language can you make such a lovely cast? :-)
		 */
		pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
							 (void *)&modify_list_mutex);
		pthread_mutex_lock(&modify_list_mutex);

		for(i = 0, nprobes = 0;
			users->last && nprobes < maxprobes && i < list_size(users);
			++i)
		{
			userdata_t *ud = (userdata_t *)users->last->data;

			/* skip users on domains we should not probe
			 */
			if(ud->domain->fwtype == FW_LOCAL &&
			   ud->domain->probelib_handle != 0)
			{
				/* Skip users already under probing
				 */
				if(get_mstatus(ud) != MS_OK) {
					list_movetofront(users, users->last);
					continue;
				}

				logmsg(LOG_DEBUG, "starting probe thread for %s@%s",
					   ud->username, ud->domain->name);

				set_mstatus(ud, MS_PROBING);

				/* Set the timeout value for this probe to
				 * probe-timeout seconds in the future.
				 */
				gettimeofday(&ud->tv, 0);
				ud->tv.tv_sec += cfg_getint(cfg, "probe-timeout");

				/* Move the user to the front of the list. We begin
				 * each probe session at the end of the list with
				 * users probed lastly. Also insert (but don't copy)
				 * the user data last in the prio list.
				 */
				list_movetofront(users, users->last);
				list_insert(prio, ud);

				/* Create a thread for the probing. We detach it so
				 * resources will be freed directly when the thread
				 * exits.
				 */
				pthread_create(&ud->probe_id, &thread_attr,
							   (void *(*)(void *))probe_thread_main,
							   ud);
				++nprobes;
			} else {
				/* move users controlled by fwcd to the front */
				list_movetofront(users, users->last);
			}
		}
		pthread_mutex_unlock(&modify_list_mutex);
		pthread_cleanup_pop(0);
	}

	/* should never be reached
	 */
	pthread_cleanup_pop(0);
	return 0;
}

int fwcd_cb(SSL *ssl)
{
	char buf[1024];
	int ret;
	fwreq_t *req;
	fwreply_t reply;
	userdata_t *ud;

	ret = SSL_read(ssl, buf, 1024);

	if(ret == 0) {
		logmsg(LOG_WARNING, "connection closed by peer");
		return -2;
	}

	if(ret < 0) {
		logmsg_ssl(LOG_ERR);
		return -1;
	}

	if(ret != sizeof(fwreq_t)) {
		logmsg(LOG_ERR, "received too small request");
		return -1;
	}

	req = (fwreq_t *)buf;

	switch(req->cmd) {
		case FWCMD_CLOSE:
			logmsg(LOG_DEBUG, "got close request for IP %s, MAC %s",
				   inet_ntoa(req->ip), strmac(req->mac));
			ud = search_ipmac(&req->ip, req->mac);
			if(ud) {
				logmsg(LOG_INFO, "logout %s@%s ip: %s, mac: %s",
					   ud->username, ud->domain->name,
					   inet_ntoa(ud->ip), strmac(ud->mac));
				reply.status = monitor_remove(ud);
			} else {
				logmsg(LOG_ERR, "close request: %s %s not logged in",
					   inet_ntoa(req->ip), strmac(req->mac));
				reply.status = REQ_STATUS_NOUSER;
			}
			break;
		case FWCMD_OPEN:
		case FWCMD_RESET:
		case FWCMD_SYNC_USER:
		case FWCMD_STATS:
		default:
			reply.status = REQ_STATUS_FAIL;
			break;
	}

	ret = SSL_write(ssl, &reply, sizeof(fwreply_t));
	if(check_ssl_io(ret, ssl, sizeof(fwreply_t)) != 0)
		return -1;

	return 0;
}

void *fwcd_thread_main(const char *domain)
{
	SSL_CTX *ctx;
	cfg_t *dsec;

	block_term_sigs();

	logmsg(LOG_INFO, "fwcd thread for domain %s starting...", domain);

	dsec = cfg_tgetsec(cfg, "domain", domain);

	ctx = init_ctx(cfg_getstr(dsec, "certificate-file"),
				   cfg_getstr(dsec, "key-file"),
				   cfg_getstr(dsec, "fwcd-certificate-file"),
				   0);

	if(ctx) {
		init_ssl_daemon(ctx, cfg_getint(dsec, "port"), fwcd_cb);

		/* FIXME: here we lost the connection to fwcd, what should we
		 * do? When fwcd eventually comes up again, we need to have
		 * the information synchorized. Easiest would be to reset the
		 * domain here (ie, remove all users from that domain) When
		 * the connection to fwcd is established again, we need to do
		 * a (remote) reset.
		 */
	}
	return 0;
}

/* initialization of the monitor
 */
void monitor_init(void)
{
	int i;
	int num_local = 0;
	int num_remote = 0;
	cfgval_t *val;

	logmsg(LOG_INFO, "initializing...");

	/* initialize users list
	 */
	users = list_new((list_delfunc_t)userdata_destroy);
	logout_users = list_new(0);

	/* initialize the domains
	 */
	domains = (domain_t *)xmalloc(num_domain * sizeof(domain_t));
	val = cfg_getval(cfg, "domain");
	for(i = 0; i < val->a.size; i++) {
		char *probe_lib;
		cfg_t *sec = &val->a.val[i].sec;

		domains[i].name = xstrdup(sec->title);
		domains[i].nusers = 0;

		if(cfg_getstr(sec, "fwcd-host") != 0) {
			domains[i].fwtype = FW_REMOTE;
			num_remote++;
			if(pthread_create(&fwcd_thread,
							  0,
							  (void *(*)(void *))fwcd_thread_main,
							  domains[i].name) != 0)
			{
				logmsg(LOG_ERR,
					   "Failed to initialize fwcd thread for domain %s",
					   domains[i].name);
				exit(7);
			}
		} else {
			domains[i].fwtype = FW_LOCAL;
			num_local++;
		}

		probe_lib = cfg_getstr(sec, "probe-library");
		domains[i].probelib_handle = 0;
		if(probe_lib && probe_lib[0]) {
			if(domains[i].fwtype == FW_REMOTE) {
				logmsg(LOG_INFO,
					   "probe-library setting ignored for remote firewall");
			} else {
				domains[i].probelib_handle = dlopen(probe_lib, RTLD_NOW);
				if(domains[i].probelib_handle == 0) {
					logmsg(LOG_ERR, "%s", dlerror());
					exit(7);
				} else {
					logmsg(LOG_INFO,
						   "registered probe library %s for domain %s",
						   probe_lib, domains[i].name);
				}
				domains[i].probe_user_func = dlsym(domains[i].probelib_handle,
												   "probe_user");
				if(domains[i].probe_user_func == 0) {
					logmsg(LOG_ERR, "probe_user(): %s",
						   dlerror());
					exit(7);
				}
				domains[i].probe_cleanup_func =
					dlsym(domains[i].probelib_handle, "probe_cleanup");
			}
		}

		/* The ssl connection will be opened when a command is sent
		 */
		domains[i].sslout = 0;
	}

	/* reset all firewalls
	 */
	reset_firewalls();

	/* initialize and start monitor thread, if there are any domains
	 * we should monitor
	 */
	if(/*num_local &&*/
	   pthread_create(&monitor_thread, 0, monitor_thread_main, 0) != 0)
	{
		logmsg(LOG_ERR, "Failed to initialize monitoring thread");
		exit(7);
	}
}

userdata_t *search_user(const char *username, const char *domain)
{
	listitem_t *li;
	userdata_t *ud;

	pthread_mutex_lock(&modify_list_mutex);

	for(li = users->first; li; li = li->next) {
		ud = (userdata_t *)li->data;
		if(strcmp(ud->username, username) == 0 &&
		   strcmp(ud->domain->name, domain) == 0)
		{
			pthread_mutex_unlock(&modify_list_mutex);
			return ud;
		}
	}
	pthread_mutex_unlock(&modify_list_mutex);

	return 0;
}

userdata_t *search_ipmac(struct in_addr *ip, unsigned char mac[6])
{
	listitem_t *li;
	userdata_t *ud;

	pthread_mutex_lock(&modify_list_mutex);

	for(li = users->first; li; li = li->next) {
		ud = (userdata_t *)li->data;
		if(ud->ip.s_addr == ip->s_addr &&
		   memcmp(ud->mac, mac, 6) == 0)
		{
			pthread_mutex_unlock(&modify_list_mutex);
			return ud;
		}
	}
	pthread_mutex_unlock(&modify_list_mutex);

	return 0;
}

mstatus_t get_mstatus(userdata_t *ud)
{
	mstatus_t mstatus;

	pthread_mutex_lock(&ud->mutex);
	mstatus = ud->mstatus;
	pthread_mutex_unlock(&ud->mutex);

	return mstatus;
}

mstatus_t set_mstatus(userdata_t *ud, mstatus_t mstatus)
{
	mstatus_t old_mstatus;

	pthread_mutex_lock(&ud->mutex);
	old_mstatus = ud->mstatus;
	ud->mstatus = mstatus;
	pthread_mutex_unlock(&ud->mutex);

	return mstatus;
}
