/* Open.Net Authentication Daemon
 * PAM Authentication
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: pamauth.c,v 1.8 2002/04/06 21:27:11 mhe Exp $
 */

#include "common.h"

#include <security/pam_appl.h>

static pthread_key_t pam_passwd_key;
static pthread_once_t pam_passwd_key_once = PTHREAD_ONCE_INIT;

static int answer(int num_msg,
				  const struct pam_message **msg,
				  struct pam_response **resp,
				  void *appdata_ptr)
{
	int i;
	char *pam_passwd;

	pam_passwd = (char *)pthread_getspecific(pam_passwd_key);
	if(pam_passwd == 0) {
		logmsg(LOG_ERR, "failed reading pam_passwd_key");
		return PAM_ABORT;
	}
	
	/* Answer all messages with the password in pam_passwd.
	 * The PAM module is responsible of free()ing the structure.
	 */

	*resp = (struct pam_response *)malloc(num_msg * sizeof(struct pam_response));
	for(i=0;i<num_msg;i++) {
		(*resp)[i].resp = (char *)malloc(strlen(pam_passwd) + 1);
		strcpy((*resp)[i].resp, pam_passwd);
	}

	return PAM_SUCCESS;
}

/* Free the thread-specific buffer */
static void pam_passwd_destroy(void *buf)
{
	free(buf);
}

/* Allocate the key */
static void pam_passwd_key_alloc()
{
	pthread_key_create(&pam_passwd_key, pam_passwd_destroy);
}

/* Allocate the thread-specific buffer */
static void pam_passwd_alloc(void)
{
	pthread_once(&pam_passwd_key_once, pam_passwd_key_alloc);
	pthread_setspecific(pam_passwd_key, malloc(1024)); /* max length of password is 1024 */
}

req_status_t authenticate(char *username, char *domain, char *password)
{
	int r;
	req_status_t status = REQ_STATUS_FAIL;
	pam_handle_t *pam_handle = 0;
	struct pam_conv pconv = { answer, 0 };
	char *service;
	char *pam_passwd = 0, *tsd;

#ifdef TEST
	return REQ_STATUS_OK;
#endif

	service = cfg_stgetstr(cfg, "domain", domain, "pam-service");
	if(!service)
		/* no service specified, this should really be checked by the
		 * configuration parser instead
		 */
		return REQ_STATUS_FAIL;

	logmsg(LOG_DEBUG, "authenticating %s@%s", username, domain);
	if(pam_start(service, username, &pconv, &pam_handle) != PAM_SUCCESS) {
		logmsg(LOG_ERR, "Failed to initialize PAM library");
		return REQ_STATUS_BACKFAIL;
	}

	pam_set_item(pam_handle, PAM_USER, username);
	pam_set_item(pam_handle, PAM_AUTHTOK, password);
	pam_set_item(pam_handle, PAM_TTY, "/dev/null");

	/* copy the password to a TSD (Thread-Specific Data) are so the
	 * answer() function above can get that information.
	 */
	pam_passwd_alloc();
	tsd = pthread_getspecific(pam_passwd_key);
	if(tsd == 0) {
		logmsg(LOG_ERR, "unable to allocate/set pam_passwd_key");
		return REQ_STATUS_FAIL;
	}
	strncpy(tsd, password, 1024);

	r = pam_authenticate(pam_handle, PAM_SILENT | PAM_DISALLOW_NULL_AUTHTOK);
	logmsg(LOG_DEBUG, "PAM says: %s", pam_strerror(pam_handle, r));
	pam_end(pam_handle, r);
	switch(r) {
		case PAM_SUCCESS:
			status = REQ_STATUS_OK;
			break;
		case PAM_ABORT:
			status = REQ_STATUS_BACKFAIL;
			break;
		case PAM_AUTHINFO_UNAVAIL:
		case PAM_AUTH_ERR:
		case PAM_USER_UNKNOWN:
		case PAM_CRED_INSUFFICIENT:
		case PAM_MAXTRIES:
		default:
			status = REQ_STATUS_FAIL;
			break;
	}
	return status;
}
