/* Open.Net Authentication Daemon
 * Socket handling and processing of requests
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: request.c,v 1.26 2002/04/06 21:27:11 mhe Exp $
 */

#include "common.h"
#include "modechange.h"
#include "xmalloc.h"
#include <grp.h>
#include <pwd.h>

#define MAX_USERNAME_LEN 42
#define MAX_DOMAIN_LEN 42
#define MAX_PASSWORD_LEN 42

struct request_t {
	req_cmd_t cmd;
	unsigned char mac[MAX_MAC_LEN];
	struct in_addr ip;
	char username[MAX_USERNAME_LEN+1];
	char domain[MAX_DOMAIN_LEN+1];
	char password[MAX_PASSWORD_LEN+1];
};
typedef struct request_t request_t;

/* some statistics counters
 */
unsigned long int failed_requests = 0;
unsigned long int failed_auths = 0;
unsigned long int total_logins = 0;
time_t oasis_start;

/* Open a UNIX socket named filename, returns file descriptor, exits
 * in case of failure
 *
 * also sets the file permission and owner/group on the socket file
 *
 * options used: socket-path, socket-owner, socket-mode
 */
static int open_socket(char *filename)
{
	int fd;
	struct sockaddr_un saddr;
	char *modestr;
	char *owner;

	unlink(filename);
	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if(fd == -1) {
		logmsg(LOG_ERR, "Unable to open socket: %s", strerror(errno));
		exit(4);
	}
	saddr.sun_family = AF_UNIX;
	strcpy(saddr.sun_path, filename);
	if(bind(fd, (struct sockaddr *)&saddr, sizeof(saddr)) != 0) {
		logmsg(LOG_ERR, "Unable to bind socket to %s: %s", filename,
			   strerror(errno));
		exit(4);
	}

	/* Set the permissions on the socket (if configured)
	 */
	modestr = cfg_getstr(cfg, "socket-mode");
	if(modestr && modestr[0]) {
		struct mode_change *cmod;
		unsigned short m;

		cmod = mode_compile(modestr, MODE_MASK_EQUALS);
		if(cmod == MODE_INVALID) {
			logmsg(LOG_ERR, "invalid mode for socket-mode: %s", modestr);
			close(fd);
			unlink(filename);
			exit(4);
		} else if(cmod == MODE_MEMORY_EXHAUSTED) {
			logmsg(LOG_ERR, "virtual memory exhausted");
			close(fd);
			unlink(filename);
			exit(4);
		}

		m = mode_adjust(0, cmod);
		logmsg(LOG_INFO, "setting socket mode to 0%o", m);
		if(chmod(filename, m) != 0)
			logmsg(LOG_ERR, "chmod(%s): %s", filename, strerror(errno));
	}

	/* Set owner and group of the socket (if configured)
	 */
	owner = cfg_getstr(cfg, "socket-owner");
	if(owner && owner[0]) {
		struct group *grp = 0;
		char *group;
		struct passwd *pwd;

		owner = strdup(owner);

		group = strchr(owner, '.');
		if(group) {
			*group++ = 0;
			grp = getgrnam(group);
			if(grp == 0) {
				logmsg(LOG_ERR, "no such group: %s", group);
				close(fd);
				unlink(filename);
				exit(4);
			}
		}

		pwd = getpwnam(owner);
		if(pwd == 0) {
			logmsg(LOG_ERR, "no such user: %s", owner);
			close(fd);
			unlink(filename);
			exit(4);
		}
		logmsg(LOG_INFO, "setting owner of socket to %s.%s", owner, group);
		chown(filename, pwd->pw_uid, grp ? grp->gr_gid : -1);

		xfree(owner);
	}

	/* now start to listen for requests on the socket
	 */
	if(listen(fd, cfg_getint(cfg, "backlog")) != 0) {
		logmsg(LOG_ERR, "Unable to listen on socket: %s", strerror(errno));
		unlink(filename);
		exit(4);
	}

	logmsg(LOG_INFO, "Listening on socket %s for auth requests...",
		   filename);

	return fd;
}

static req_ret_t decode_request(char *str, request_t *req)
{
	char *e;

	e = strsep(&str, " \t\r\n");
	if(strcasecmp(e, "LOGIN") == 0)
		req->cmd = REQ_CMD_LOGIN;
	else if(strcasecmp(e, "LOGOUT") == 0)
		req->cmd = REQ_CMD_LOGOUT;
	else if(strcasecmp(e, "STATS") == 0)
		req->cmd = REQ_CMD_STATS;
	else if(strcasecmp(e, "LIST") == 0)
		req->cmd = REQ_CMD_LIST;
	else {
		logmsg(LOG_WARNING, "unknown request command: %s", e);
		failed_requests++;
		return REQ_RET_FAIL;
	}

	while(1) {
		char *opt;
		char *val;

		if(str == 0)
			break;

		opt = strsep(&str, "=\r\n");
		val = strsep(&str, "\r\n");

		if(!opt || !val)
			return REQ_RET_FAIL;
		if(strcasecmp(opt, "USER") == 0)
			xstrncpy(req->username, val, MAX_USERNAME_LEN);
		else if(strcasecmp(opt, "DOMAIN") == 0)
			xstrncpy(req->domain, val, MAX_DOMAIN_LEN);
		else if(strcasecmp(opt, "PASS") == 0)
			xstrncpy(req->password, val, MAX_PASSWORD_LEN);
		else if(strcasecmp(opt, "IP") == 0) {
			if(inet_aton(val, &req->ip) == 0)
				return REQ_RET_FAIL;
		} else if(strcasecmp(opt, "MAC") == 0) {
			int tmp[6];
			int i;

			if(sscanf(val, "%2x:%2x:%2x:%2x:%2x:%2x",
					  &tmp[0], &tmp[1], &tmp[2],
					  &tmp[3], &tmp[4], &tmp[5]) != 6)
				return REQ_RET_FAIL;

			for(i = 0; i < 6; i++)
				req->mac[i] = (char)(tmp[i] & 0xFF);

		} else {
			logmsg(LOG_WARNING, "unknown request option: %s", opt);
			failed_requests++;
			return REQ_RET_FAIL;
		}
	}

	return REQ_RET_OK;
}

/* Read and process a request from file descriptor fd
 */
static void process_request(int fd)
{
	request_t req;
	char *reqstr;
	struct sockaddr_un caddr;
	socklen_t clen;
	int cfd;
	int send_stats = 0;
	int send_list = 0;
	userdata_t *ud;
	char buf[1025];
	FILE *crfp, *cwfp;
	req_ret_t ret;
	req_status_t status = REQ_STATUS_FAIL;

	clen = sizeof(caddr);
	cfd = accept(fd, (struct sockaddr *)&caddr, &clen);
	if(cfd == -1) {
		logmsg(LOG_ERR, "Couldn't accept connection: %s", strerror(errno));
		return;
	}
	crfp = fdopen(cfd, "r");
	cwfp = fdopen(cfd, "w");

	if(fgets(buf, 1024, crfp) == 0) {
		logmsg(LOG_WARNING, "read(): %s", strerror(errno));
		fclose(crfp);
		fclose(cwfp);
		close(cfd);
		return;
	}
	reqstr = trim(buf);

	memset(&req, 0, sizeof(request_t));
	ret = decode_request(reqstr, &req);

	if(ret == REQ_RET_OK) {
		switch(req.cmd) {
			case REQ_CMD_LOGIN:
				logmsg(LOG_DEBUG, "got login request for %s@%s",
					   req.username, req.domain);
				total_logins++;
				if(req.domain[0] == 0 || req.username[0] == 0 ||
				   req.password[0] == 0)
					status = REQ_STATUS_BADREQ;
				else {
					ud = userdata_create(req.username, req.domain,
										 req.password, &req.ip,
										 req.mac);
					status = login(ud, req.password);
					if(status != REQ_STATUS_OK) {
						logmsg(LOG_INFO,
							   "login failed for %s@%s (%s)",
							   req.username, req.domain, strstatus(status));
					}
					if(status != REQ_STATUS_OK &&
					   status != REQ_STATUS_FWFAILED &&
					   status != REQ_STATUS_FWCONNFAIL)
						userdata_destroy(ud);
				}
				break;
			case REQ_CMD_LOGOUT:
				logmsg(LOG_INFO, "got logout request for %s@%s",
					   req.username, req.domain);

				ud = search_user(req.username, req.domain);

				if(!ud) {
					status = REQ_STATUS_NOUSER;
					userdata_destroy(ud);
				} else {
					if(ud->mstatus == MS_LOGOUT || ud->mstatus == MS_DETACHED)
					{
						logmsg(LOG_INFO, "user %s@%s already being logged out",
							   ud->username, ud->domain->name);
						status = REQ_STATUS_OK;
					} else {
						pthread_mutex_lock(&logout_users_mutex);
						logmsg(LOG_DEBUG, "queuing logout request for %s@%s",
							   ud->username, ud->domain->name);
						list_insert(logout_users, ud);
						pthread_mutex_unlock(&logout_users_mutex);
						status = REQ_STATUS_OK;
					}
				}
				break;
			case REQ_CMD_STATS:
				logmsg(LOG_DEBUG, "got stats request");
				send_stats = 1;
				status = REQ_STATUS_OK;
				break;
			case REQ_CMD_LIST:
				logmsg(LOG_DEBUG, "got userlist request");
				send_list = 1;
				status = REQ_STATUS_OK;
				break;
			default:
				logmsg(LOG_WARNING, "got unknown request");
				ret = REQ_RET_FAIL;
				status = REQ_STATUS_NOCMD;
		}
	} else {
		failed_requests++;
		status = REQ_STATUS_BADREQ;
	}

	fprintf(cwfp, "%d %s\r\n", status, strstatus(status));
	/* Ignore errors: if the user is logged in properly, but we failed
	 * to write back the status, just ignore the error and let the
	 * user stay logged in
	 */

	if(send_list) {
		if(req.domain[0]) {
			if(req.username[0])
				stats_send_user(cwfp, req.username, req.domain);
#if 0
			else
				stats_send_domain(cwfp, req.domain);
#endif
		} else
			stats_send(cwfp);
	} else if(send_stats) {
		pthread_mutex_lock(&modify_list_mutex);
		fprintf(cwfp, "%u %lu %lu %lu %lu\r\n",
				list_size(users),
				total_logins,
				failed_auths,
				failed_requests,
				time(0) - oasis_start);
		pthread_mutex_unlock(&modify_list_mutex);
	}
	fflush(cwfp);

	fclose(crfp);
	fclose(cwfp);
	close(cfd);
}

/* Wait for incoming requests on the socket
 */
void request_loop(void)
{
	int sockfd = open_socket(cfg_getstr(cfg, "socket-path"));

	while(1) {
		fd_set fds;
		int n;

		FD_ZERO(&fds);
		FD_SET(sockfd, &fds);
		n = select(sockfd+1, &fds, 0, 0, 0);
		if(n == -1)
			logmsg(LOG_WARNING, "select() failed: %s", strerror(errno));
		else if(FD_ISSET(sockfd, &fds))
			process_request(sockfd);
	}
}
